from hummingbot.core.api_throttler.data_types import RateLimit
from hummingbot.core.data_type.in_flight_order import OrderState

DEFAULT_DOMAIN = "com.br"

HBOT_ORDER_ID_PREFIX = "55"
USER_AGENT = "HBOT"
MAX_ORDER_ID_LEN = 20

# Base URL
REST_URL = "https://api.foxbit.com.br"
REST_V2_URL = "https://api.foxbit.com.br/AP"
WSS_URL = "api.foxbit.com.br"

PUBLIC_API_VERSION = "v3"
PRIVATE_API_VERSION = "v3"

# Public API endpoints or FoxbitClient function
EXCHANGE_INFO_PATH_URL = "markets"
PING_PATH_URL = "system/time"
SNAPSHOT_PATH_URL = "markets/{}/orderbook"
SERVER_TIME_PATH_URL = "system/time"
INSTRUMENTS_PATH_URL = "GetInstruments"

# Private API endpoints or FoxbitClient function
ACCOUNTS_PATH_URL = "accounts"
MY_TRADES_PATH_URL = "trades"
ORDER_PATH_URL = "orders"
CANCEL_ORDER_PATH_URL = "orders/cancel"
GET_ORDER_BY_CLIENT_ID = "orders/by-client-order-id/{}"

WS_HEADER = {
    "Content-Type": "application/json",
    "User-Agent": USER_AGENT,
}

WS_MESSAGE_FRAME_TYPE = {
    "Request": 0,
    "Reply": 1,
    "Subscribe": 2,
    "Event": 3,
    "Unsubscribe": 4,
}

WS_MESSAGE_FRAME = {
    "m": 0,  # WS_MESSAGE_FRAME_TYPE
    "i": 0,  # Sequence Number
    "n": "",  # Endpoint
    "o": "",  # Message Payload
}

WS_HEARTBEAT_TIME_INTERVAL = 20

SIDE_BUY = 'BUY'
SIDE_SELL = 'SELL'

# Rate Limit time intervals
ONE_MINUTE = 60
ONE_SECOND = 1
TWO_SECONDS = 2
ONE_DAY = 86400

# Order States
ORDER_STATE = {
    "PENDING": OrderState.PENDING_CREATE,
    "ACTIVE": OrderState.OPEN,
    "NEW": OrderState.OPEN,
    "FILLED": OrderState.FILLED,
    "PARTIALLY_FILLED": OrderState.PARTIALLY_FILLED,
    "PENDING_CANCEL": OrderState.PENDING_CANCEL,
    "CANCELED": OrderState.CANCELED,
    "PARTIALLY_CANCELED": OrderState.CANCELED,
    "REJECTED": OrderState.FAILED,
    "EXPIRED": OrderState.FAILED,
    "Unknown": OrderState.PENDING_CREATE,
    "Working": OrderState.OPEN,
    "Rejected": OrderState.FAILED,
    "Canceled": OrderState.CANCELED,
    "Expired": OrderState.FAILED,
    "FullyExecuted": OrderState.FILLED,
}

# Websocket subscribe endpoint
WS_AUTHENTICATE_USER = "AuthenticateUser"
WS_SUBSCRIBE_ACCOUNT = "SubscribeAccountEvents"
WS_SUBSCRIBE_ORDER_BOOK = "SubscribeLevel2"
WS_SUBSCRIBE_TOB = "SubscribeLevel1"
WS_SUBSCRIBE_TRADES = "SubscribeTrades"

# Websocket response event types from Foxbit
# Market data events
WS_ORDER_BOOK_RESPONSE = "Level2UpdateEvent"
# Private order events
WS_ACCOUNT_POSITION = "AccountPositionEvent"
WS_ORDER_STATE = "OrderStateEvent"
WS_ORDER_TRADE = "OrderTradeEvent"
WS_TRADE_RESPONSE = "TradeDataUpdateEvent"

ORDER_BOOK_DEPTH = 10

RATE_LIMITS = [
    RateLimit(limit_id=EXCHANGE_INFO_PATH_URL, limit=6, time_interval=ONE_SECOND),
    RateLimit(limit_id=SNAPSHOT_PATH_URL, limit=10, time_interval=TWO_SECONDS),
    RateLimit(limit_id=SERVER_TIME_PATH_URL, limit=5, time_interval=ONE_SECOND),
    RateLimit(limit_id=PING_PATH_URL, limit=5, time_interval=ONE_SECOND),
    RateLimit(limit_id=ACCOUNTS_PATH_URL, limit=15, time_interval=ONE_SECOND),
    RateLimit(limit_id=MY_TRADES_PATH_URL, limit=5, time_interval=ONE_SECOND),
    RateLimit(limit_id=GET_ORDER_BY_CLIENT_ID, limit=30, time_interval=TWO_SECONDS),
    RateLimit(limit_id=CANCEL_ORDER_PATH_URL, limit=30, time_interval=TWO_SECONDS),
    RateLimit(limit_id=ORDER_PATH_URL, limit=30, time_interval=TWO_SECONDS),
    RateLimit(limit_id=INSTRUMENTS_PATH_URL, limit=750, time_interval=ONE_MINUTE),
]

# Error codes
ORDER_NOT_EXIST_MESSAGE = "HTTP status is 404"



import hashlib
import hmac
import json
import time
from typing import Any, Dict
from urllib.parse import urlparse

import eth_account
from eth_account.messages import encode_typed_data
from eth_utils import to_hex
from yarl import URL

from hummingbot.connector.derivative.aevo_perpetual import aevo_perpetual_constants as CONSTANTS
from hummingbot.core.web_assistant.auth import AuthBase
from hummingbot.core.web_assistant.connections.data_types import RESTMethod, RESTRequest, WSRequest


class AevoPerpetualAuth(AuthBase):
    def __init__(
        self,
        api_key: str,
        api_secret: str,
        signing_key: str,
        account_address: str,
        domain: str,
    ):
        self._api_key = api_key
        self._api_secret = api_secret
        self._signing_key = signing_key
        self._account_address = account_address
        self._domain = domain
        self._wallet = eth_account.Account.from_key(signing_key)

    @property
    def api_key(self) -> str:
        return self._api_key

    @property
    def api_secret(self) -> str:
        return self._api_secret

    @property
    def account_address(self) -> str:
        return self._account_address

    async def rest_authenticate(self, request: RESTRequest) -> RESTRequest:
        timestamp = str(int(time.time() * 1e9))
        if request.url is None:
            raise ValueError("Request URL is required for Aevo authentication.")
        parsed_url = urlparse(request.url)
        path = parsed_url.path
        body = ""

        if request.method in [RESTMethod.GET, RESTMethod.DELETE] and request.params:
            sorted_params = [(str(k), str(v)) for k, v in sorted(request.params.items(), key=lambda item: item[0])]
            request.params = sorted_params
        elif parsed_url.query:
            path = URL(request.url).raw_path_qs
        elif request.data is not None:
            if isinstance(request.data, (dict, list)):
                request.data = json.dumps(request.data)
            else:
                body = str(request.data)

        payload = f"{self._api_key},{timestamp},{request.method.value.upper()},{path},{body}"
        signature = hmac.new(
            self._api_secret.encode("utf-8"),
            payload.encode("utf-8"),
            hashlib.sha256,
        ).hexdigest()

        headers = request.headers or {}
        headers.update({
            "AEVO-TIMESTAMP": timestamp,
            "AEVO-SIGNATURE": signature,
            "AEVO-KEY": self._api_key,
        })
        request.headers = headers

        return request

    async def ws_authenticate(self, request: WSRequest) -> WSRequest:
        return request

    def get_ws_auth_payload(self) -> Dict[str, Any]:
        return {
            "op": "auth",
            "data": {
                "key": self._api_key,
                "secret": self._api_secret,
            },
        }

    def sign_order(
        self,
        is_buy: bool,
        limit_price: int,
        amount: int,
        salt: int,
        instrument: int,
        timestamp: int,
    ) -> str:
        domain = {
            "name": "Aevo Mainnet" if self._domain == CONSTANTS.DEFAULT_DOMAIN else "Aevo Testnet",
            "version": "1",
            "chainId": 1 if self._domain == CONSTANTS.DEFAULT_DOMAIN else 11155111,
        }
        types = {
            "EIP712Domain": [
                {"name": "name", "type": "string"},
                {"name": "version", "type": "string"},
                {"name": "chainId", "type": "uint256"},
            ],
            "Order": [
                {"name": "maker", "type": "address"},
                {"name": "isBuy", "type": "bool"},
                {"name": "limitPrice", "type": "uint256"},
                {"name": "amount", "type": "uint256"},
                {"name": "salt", "type": "uint256"},
                {"name": "instrument", "type": "uint256"},
                {"name": "timestamp", "type": "uint256"},
            ],
        }
        message = {
            "maker": self._account_address,
            "isBuy": is_buy,
            "limitPrice": int(limit_price),
            "amount": int(amount),
            "salt": int(salt),
            "instrument": int(instrument),
            "timestamp": int(timestamp),
        }
        typed_data = {
            "domain": domain,
            "types": types,
            "primaryType": "Order",
            "message": message,
        }
        encoded = encode_typed_data(full_message=typed_data)
        signed = self._wallet.sign_message(encoded)

        return to_hex(signed.signature)



import asyncio
import json
import logging
import platform
from abc import ABC, abstractmethod
from decimal import Decimal
from os.path import dirname, join, realpath
from typing import TYPE_CHECKING, List, Tuple

from hummingbot.connector.utils import combine_to_hb_trading_pair, split_hb_trading_pair
from hummingbot.core.event.event_forwarder import EventForwarder
from hummingbot.core.event.events import MarketEvent, OrderFilledEvent
from hummingbot.core.py_time_iterator import PyTimeIterator
from hummingbot.core.rate_oracle.rate_oracle import RateOracle
from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.logger import HummingbotLogger
from hummingbot.logger.log_server_client import LogServerClient

if TYPE_CHECKING:
    from hummingbot.connector.connector_base import ConnectorBase

with open(realpath(join(dirname(__file__), '../VERSION'))) as version_file:
    CLIENT_VERSION = version_file.read().strip()


class MetricsCollector(PyTimeIterator, ABC):
    """Base class for metrics collectors"""

    DEFAULT_METRICS_SERVER_URL = "https://api.coinalpha.com/reporting-proxy-v2"

    @abstractmethod
    def start(self):
        raise NotImplementedError

    @abstractmethod
    def stop(self):
        raise NotImplementedError

    def tick(self, timestamp: float):
        pass


class DummyMetricsCollector(MetricsCollector):

    def start(self):
        # Nothing is required
        pass

    def stop(self):
        # Nothing is required
        pass


class TradeVolumeMetricCollector(MetricsCollector):

    _logger = None

    METRIC_NAME = "filled_usdt_volume"

    def __init__(self,
                 connector: 'ConnectorBase',
                 activation_interval: Decimal,
                 rate_provider: RateOracle,
                 instance_id: str,
                 valuation_token: str = "USDT"):
        super().__init__()
        self._connector = connector
        self._activation_interval = activation_interval
        self._dispatcher = LogServerClient(log_server_url=self.DEFAULT_METRICS_SERVER_URL)
        self._rate_provider = rate_provider
        self._instance_id = instance_id
        self._client_version = CLIENT_VERSION
        self._valuation_token = valuation_token
        self._last_process_tick_timestamp = 0
        self._last_executed_collection_process = None
        self._collected_events = []

        self._fill_event_forwarder = EventForwarder(self._register_fill_event)

        self._event_pairs: List[Tuple[MarketEvent, EventForwarder]] = [
            (MarketEvent.OrderFilled, self._fill_event_forwarder),
        ]

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls._logger is None:
            cls._logger = logging.getLogger(__name__)
        return cls._logger

    def start(self):
        self.register_listener()
        self._dispatcher.start()

    def stop(self):
        self.trigger_metrics_collection_process()
        for event_pair in self._event_pairs:
            self._connector.remove_listener(event_pair[0], event_pair[1])
        self._dispatcher.stop()

    def register_listener(self):
        for event_pair in self._event_pairs:
            self._connector.add_listener(event_pair[0], event_pair[1])

    def tick(self, timestamp: float):
        if self._fill_event_forwarder not in self._connector.get_listeners(MarketEvent.OrderFilled):
            self.register_listener()
        inactivity_time = timestamp - self._last_process_tick_timestamp
        if inactivity_time >= self._activation_interval:
            self._last_process_tick_timestamp = timestamp
            self.trigger_metrics_collection_process()

    def trigger_metrics_collection_process(self):
        events_to_process = self._collected_events
        self._collected_events = []
        self._last_executed_collection_process = safe_ensure_future(
            self.collect_metrics(events=events_to_process))

    async def collect_metrics(self, events: List[OrderFilledEvent]):
        try:
            total_volume = Decimal("0")

            for fill_event in events:
                trade_base, trade_quote = split_hb_trading_pair(fill_event.trading_pair)
                from_quote_conversion_pair = combine_to_hb_trading_pair(base=trade_quote, quote=self._valuation_token)
                rate = await self._rate_provider.stored_or_live_rate(from_quote_conversion_pair)

                if rate is not None:
                    total_volume += fill_event.amount * fill_event.price * rate
                else:
                    from_base_conversion_pair = combine_to_hb_trading_pair(base=trade_base, quote=self._valuation_token)
                    rate = await self._rate_provider.stored_or_live_rate(from_base_conversion_pair)
                    if rate is not None:
                        total_volume += fill_event.amount * rate
                    else:
                        self.logger().debug(f"Could not find a conversion rate rate using Rate Oracle for any of "
                                            f"the pairs {from_quote_conversion_pair} or {from_base_conversion_pair}")

            if total_volume > Decimal("0"):
                self._dispatch_trade_volume(total_volume)
        except asyncio.CancelledError:
            raise
        except Exception:
            self._collected_events.extend(events)

    def _dispatch_trade_volume(self, volume: Decimal):
        metric_request = {
            "url": f"{self._dispatcher.log_server_url}/client_metrics",
            "method": "POST",
            "request_obj": {
                "headers": {
                    'Content-Type': "application/json"
                },
                "data": json.dumps({
                    "source": "hummingbot",
                    "name": self.METRIC_NAME,
                    "instance_id": self._instance_id,
                    "exchange": self._connector.name,
                    "version": self._client_version,
                    "system": f"{platform.system()} {platform.release()}({platform.platform()})",
                    "value": str(volume)}),
                "params": {"ddtags": f"instance_id:{self._instance_id},"
                                     f"client_version:{self._client_version},"
                                     f"type:metrics",
                           "ddsource": "hummingbot-client"}
            }
        }

        self._dispatcher.request(metric_request)

    def _register_fill_event(self, event: OrderFilledEvent):
        self._collected_events.append(event)



import asyncio
from collections import defaultdict
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from hummingbot.connector.exchange.vertex import (
    vertex_constants as CONSTANTS,
    vertex_utils as utils,
    vertex_web_utils as web_utils,
)
from hummingbot.connector.exchange.vertex.vertex_order_book import VertexOrderBook
from hummingbot.core.api_throttler.async_throttler import AsyncThrottler
from hummingbot.core.data_type.order_book_message import OrderBookMessage
from hummingbot.core.data_type.order_book_tracker_data_source import OrderBookTrackerDataSource
from hummingbot.core.web_assistant.connections.data_types import RESTMethod, WSJSONRequest
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.core.web_assistant.ws_assistant import WSAssistant

if TYPE_CHECKING:
    from hummingbot.connector.exchange.vertex.vertex_exchange import VertexExchange


class VertexAPIOrderBookDataSource(OrderBookTrackerDataSource):
    _DYNAMIC_SUBSCRIBE_ID_START = 100
    _next_subscribe_id: int = _DYNAMIC_SUBSCRIBE_ID_START

    def __init__(
        self,
        trading_pairs: List[str],
        connector: "VertexExchange",
        api_factory: Optional[WebAssistantsFactory] = None,
        domain: str = CONSTANTS.DEFAULT_DOMAIN,
        throttler: Optional[AsyncThrottler] = None,
    ):
        super().__init__(trading_pairs)
        self._connector = connector
        self._domain = domain
        self._throttler = throttler
        self._api_factory = api_factory or web_utils.build_api_factory(
            throttler=self._throttler,
        )
        self._message_queue: Dict[str, asyncio.Queue] = defaultdict(asyncio.Queue)
        self._last_ws_message_sent_timestamp = 0
        self._ping_interval = 0

    async def get_last_traded_prices(self, trading_pairs: List[str], domain: Optional[str] = None) -> Dict[str, float]:
        return await self._connector.get_last_traded_prices(trading_pairs=trading_pairs)

    async def _order_book_snapshot(self, trading_pair: str) -> OrderBookMessage:
        snapshot = await self._request_order_book_snapshot(trading_pair)
        snapshot_timestamp = utils.convert_timestamp(snapshot["data"]["timestamp"])
        snapshot_msg: OrderBookMessage = VertexOrderBook.snapshot_message_from_exchange_rest(
            snapshot, snapshot_timestamp, metadata={"trading_pair": trading_pair}
        )
        return snapshot_msg

    async def _request_order_book_snapshot(self, trading_pair: str) -> Dict[str, Any]:
        """
        Retrieves a copy of the full order book from the exchange, for a particular trading pair.

        :param trading_pair: the trading pair for which the order book will be retrieved

        :return: the response from the exchange (JSON dictionary)
        """
        product_id = utils.trading_pair_to_product_id(trading_pair, self._connector._exchange_market_info[self._domain])
        params = {
            "type": CONSTANTS.MARKET_LIQUIDITY_REQUEST_TYPE,
            "product_id": product_id,
            "depth": CONSTANTS.ORDER_BOOK_DEPTH,
        }
        rest_assistant = await self._api_factory.get_rest_assistant()

        data = await rest_assistant.execute_request(
            url=web_utils.public_rest_url(path_url=CONSTANTS.QUERY_PATH_URL, domain=self._domain),
            params=params,
            method=RESTMethod.GET,
            throttler_limit_id=CONSTANTS.MARKET_LIQUIDITY_REQUEST_TYPE,
        )
        return data

    async def _parse_trade_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        trading_pair = utils.market_to_trading_pair(
            self._connector._exchange_market_info[self._domain][raw_message["product_id"]]["market"]
        )
        metadata = {"trading_pair": trading_pair}
        trade_message: OrderBookMessage = VertexOrderBook.trade_message_from_exchange(raw_message, metadata=metadata)
        message_queue.put_nowait(trade_message)

    async def _parse_order_book_diff_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        trading_pair = utils.market_to_trading_pair(
            self._connector._exchange_market_info[self._domain][raw_message["product_id"]]["market"]
        )
        metadata = {"trading_pair": trading_pair}
        order_book_message: OrderBookMessage = VertexOrderBook.diff_message_from_exchange(
            raw_message, metadata=metadata
        )
        message_queue.put_nowait(order_book_message)

    async def _subscribe_channels(self, websocket_assistant: WSAssistant):
        """
        Subscribes to the trade events and diff orders events through the provided websocket connection.

        :param websocket_assistant: the websocket assistant used to connect to the exchange
        """

        try:
            for trading_pair in self._trading_pairs:
                product_id = utils.trading_pair_to_product_id(
                    trading_pair, self._connector._exchange_market_info[self._domain]
                )
                trade_payload = {
                    "method": CONSTANTS.WS_SUBSCRIBE_METHOD,
                    "stream": {"type": CONSTANTS.TRADE_EVENT_TYPE, "product_id": product_id},
                    "id": product_id,
                }
                subscribe_trade_request: WSJSONRequest = WSJSONRequest(payload=trade_payload)

                order_book_payload = {
                    "method": CONSTANTS.WS_SUBSCRIBE_METHOD,
                    "stream": {"type": CONSTANTS.DIFF_EVENT_TYPE, "product_id": product_id},
                    "id": product_id,
                }
                subscribe_order_book_dif_request: WSJSONRequest = WSJSONRequest(payload=order_book_payload)

                await websocket_assistant.send(subscribe_trade_request)
                await websocket_assistant.send(subscribe_order_book_dif_request)

                self._last_ws_message_sent_timestamp = self._time()

                self.logger().info(f"Subscribed to public trade and order book diff channels of {trading_pair}...")
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().error(
                "Unexpected error occurred subscribing to trading and order book stream...", exc_info=True
            )
            raise

    def _channel_originating_message(self, event_message: Dict[str, Any]) -> str:
        channel = ""

        if "type" in event_message:
            event_channel = event_message.get("type")
            if event_channel == CONSTANTS.TRADE_EVENT_TYPE:
                channel = self._trade_messages_queue_key
            if event_channel == CONSTANTS.DIFF_EVENT_TYPE:
                channel = self._diff_messages_queue_key

        return channel

    async def _process_websocket_messages(self, websocket_assistant: WSAssistant):
        """
        Connects to the trade events and order diffs websocket endpoints and listens to the messages sent by the
        exchange. Each message is stored in its own queue.
        """
        while True:
            try:
                seconds_until_next_ping = self._ping_interval - (self._time() - self._last_ws_message_sent_timestamp)

                await asyncio.wait_for(
                    super()._process_websocket_messages(websocket_assistant=websocket_assistant),
                    timeout=seconds_until_next_ping,
                )
            except asyncio.TimeoutError:
                ping_time = self._time()
                await websocket_assistant.ping()
                self._last_ws_message_sent_timestamp = ping_time

    async def _connected_websocket_assistant(self) -> WSAssistant:
        ws_url = f"{CONSTANTS.WS_SUBSCRIBE_URLS[self._domain]}"

        self._ping_interval = CONSTANTS.HEARTBEAT_TIME_INTERVAL

        websocket_assistant: WSAssistant = await self._api_factory.get_ws_assistant()

        await websocket_assistant.connect(ws_url=ws_url, message_timeout=self._ping_interval)

        return websocket_assistant

    @classmethod
    def _get_next_subscribe_id(cls) -> int:
        subscribe_id = cls._next_subscribe_id
        cls._next_subscribe_id += 1
        return subscribe_id

    async def subscribe_to_trading_pair(self, trading_pair: str) -> bool:
        """
        Subscribe to order book and trade channels for a single trading pair.

        :param trading_pair: the trading pair to subscribe to
        :return: True if successful, False otherwise
        """
        if self._ws_assistant is None:
            self.logger().warning("Cannot subscribe: WebSocket connection not established")
            return False

        try:
            product_id = utils.trading_pair_to_product_id(
                trading_pair, self._connector._exchange_market_info[self._domain]
            )
            trade_payload = {
                "method": CONSTANTS.WS_SUBSCRIBE_METHOD,
                "stream": {"type": CONSTANTS.TRADE_EVENT_TYPE, "product_id": product_id},
                "id": product_id,
            }
            subscribe_trade_request: WSJSONRequest = WSJSONRequest(payload=trade_payload)

            order_book_payload = {
                "method": CONSTANTS.WS_SUBSCRIBE_METHOD,
                "stream": {"type": CONSTANTS.DIFF_EVENT_TYPE, "product_id": product_id},
                "id": product_id,
            }
            subscribe_order_book_dif_request: WSJSONRequest = WSJSONRequest(payload=order_book_payload)

            await self._ws_assistant.send(subscribe_trade_request)
            await self._ws_assistant.send(subscribe_order_book_dif_request)

            self._last_ws_message_sent_timestamp = self._time()

            self.add_trading_pair(trading_pair)
            self.logger().info(f"Subscribed to public trade and order book diff channels of {trading_pair}...")
            return True
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().error(
                f"Unexpected error occurred subscribing to {trading_pair}...",
                exc_info=True
            )
            return False

    async def unsubscribe_from_trading_pair(self, trading_pair: str) -> bool:
        """
        Unsubscribe from order book and trade channels for a single trading pair.

        :param trading_pair: the trading pair to unsubscribe from
        :return: True if successful, False otherwise
        """
        if self._ws_assistant is None:
            self.logger().warning("Cannot unsubscribe: WebSocket connection not established")
            return False

        try:
            product_id = utils.trading_pair_to_product_id(
                trading_pair, self._connector._exchange_market_info[self._domain]
            )
            trade_payload = {
                "method": CONSTANTS.WS_UNSUBSCRIBE_METHOD,
                "stream": {"type": CONSTANTS.TRADE_EVENT_TYPE, "product_id": product_id},
                "id": product_id,
            }
            unsubscribe_trade_request: WSJSONRequest = WSJSONRequest(payload=trade_payload)

            order_book_payload = {
                "method": CONSTANTS.WS_UNSUBSCRIBE_METHOD,
                "stream": {"type": CONSTANTS.DIFF_EVENT_TYPE, "product_id": product_id},
                "id": product_id,
            }
            unsubscribe_order_book_dif_request: WSJSONRequest = WSJSONRequest(payload=order_book_payload)

            await self._ws_assistant.send(unsubscribe_trade_request)
            await self._ws_assistant.send(unsubscribe_order_book_dif_request)

            self._last_ws_message_sent_timestamp = self._time()

            self.remove_trading_pair(trading_pair)
            self.logger().info(f"Unsubscribed from public trade and order book diff channels of {trading_pair}...")
            return True
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().error(
                f"Unexpected error occurred unsubscribing from {trading_pair}...",
                exc_info=True
            )
            return False



from hummingbot.connector.exchange.mexc.protobuf import PublicDealsV3Api_pb2 as _PublicDealsV3Api_pb2
from hummingbot.connector.exchange.mexc.protobuf import PublicIncreaseDepthsV3Api_pb2 as _PublicIncreaseDepthsV3Api_pb2
from hummingbot.connector.exchange.mexc.protobuf import PublicLimitDepthsV3Api_pb2 as _PublicLimitDepthsV3Api_pb2
from hummingbot.connector.exchange.mexc.protobuf import PrivateOrdersV3Api_pb2 as _PrivateOrdersV3Api_pb2
from hummingbot.connector.exchange.mexc.protobuf import PublicBookTickerV3Api_pb2 as _PublicBookTickerV3Api_pb2
from hummingbot.connector.exchange.mexc.protobuf import PrivateDealsV3Api_pb2 as _PrivateDealsV3Api_pb2
from hummingbot.connector.exchange.mexc.protobuf import PrivateAccountV3Api_pb2 as _PrivateAccountV3Api_pb2
from hummingbot.connector.exchange.mexc.protobuf import PublicSpotKlineV3Api_pb2 as _PublicSpotKlineV3Api_pb2
from hummingbot.connector.exchange.mexc.protobuf import PublicMiniTickerV3Api_pb2 as _PublicMiniTickerV3Api_pb2
from hummingbot.connector.exchange.mexc.protobuf import PublicMiniTickersV3Api_pb2 as _PublicMiniTickersV3Api_pb2
from hummingbot.connector.exchange.mexc.protobuf import PublicBookTickerBatchV3Api_pb2 as _PublicBookTickerBatchV3Api_pb2
from hummingbot.connector.exchange.mexc.protobuf import PublicIncreaseDepthsBatchV3Api_pb2 as _PublicIncreaseDepthsBatchV3Api_pb2
from hummingbot.connector.exchange.mexc.protobuf import PublicAggreDepthsV3Api_pb2 as _PublicAggreDepthsV3Api_pb2
from hummingbot.connector.exchange.mexc.protobuf import PublicAggreDealsV3Api_pb2 as _PublicAggreDealsV3Api_pb2
from hummingbot.connector.exchange.mexc.protobuf import PublicAggreBookTickerV3Api_pb2 as _PublicAggreBookTickerV3Api_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class PushDataV3ApiWrapper(_message.Message):
    __slots__ = ("channel", "publicDeals", "publicIncreaseDepths", "publicLimitDepths", "privateOrders", "publicBookTicker", "privateDeals", "privateAccount", "publicSpotKline", "publicMiniTicker", "publicMiniTickers", "publicBookTickerBatch", "publicIncreaseDepthsBatch", "publicAggreDepths", "publicAggreDeals", "publicAggreBookTicker", "symbol", "symbolId", "createTime", "sendTime")
    CHANNEL_FIELD_NUMBER: _ClassVar[int]
    PUBLICDEALS_FIELD_NUMBER: _ClassVar[int]
    PUBLICINCREASEDEPTHS_FIELD_NUMBER: _ClassVar[int]
    PUBLICLIMITDEPTHS_FIELD_NUMBER: _ClassVar[int]
    PRIVATEORDERS_FIELD_NUMBER: _ClassVar[int]
    PUBLICBOOKTICKER_FIELD_NUMBER: _ClassVar[int]
    PRIVATEDEALS_FIELD_NUMBER: _ClassVar[int]
    PRIVATEACCOUNT_FIELD_NUMBER: _ClassVar[int]
    PUBLICSPOTKLINE_FIELD_NUMBER: _ClassVar[int]
    PUBLICMINITICKER_FIELD_NUMBER: _ClassVar[int]
    PUBLICMINITICKERS_FIELD_NUMBER: _ClassVar[int]
    PUBLICBOOKTICKERBATCH_FIELD_NUMBER: _ClassVar[int]
    PUBLICINCREASEDEPTHSBATCH_FIELD_NUMBER: _ClassVar[int]
    PUBLICAGGREDEPTHS_FIELD_NUMBER: _ClassVar[int]
    PUBLICAGGREDEALS_FIELD_NUMBER: _ClassVar[int]
    PUBLICAGGREBOOKTICKER_FIELD_NUMBER: _ClassVar[int]
    SYMBOL_FIELD_NUMBER: _ClassVar[int]
    SYMBOLID_FIELD_NUMBER: _ClassVar[int]
    CREATETIME_FIELD_NUMBER: _ClassVar[int]
    SENDTIME_FIELD_NUMBER: _ClassVar[int]
    channel: str
    publicDeals: _PublicDealsV3Api_pb2.PublicDealsV3Api
    publicIncreaseDepths: _PublicIncreaseDepthsV3Api_pb2.PublicIncreaseDepthsV3Api
    publicLimitDepths: _PublicLimitDepthsV3Api_pb2.PublicLimitDepthsV3Api
    privateOrders: _PrivateOrdersV3Api_pb2.PrivateOrdersV3Api
    publicBookTicker: _PublicBookTickerV3Api_pb2.PublicBookTickerV3Api
    privateDeals: _PrivateDealsV3Api_pb2.PrivateDealsV3Api
    privateAccount: _PrivateAccountV3Api_pb2.PrivateAccountV3Api
    publicSpotKline: _PublicSpotKlineV3Api_pb2.PublicSpotKlineV3Api
    publicMiniTicker: _PublicMiniTickerV3Api_pb2.PublicMiniTickerV3Api
    publicMiniTickers: _PublicMiniTickersV3Api_pb2.PublicMiniTickersV3Api
    publicBookTickerBatch: _PublicBookTickerBatchV3Api_pb2.PublicBookTickerBatchV3Api
    publicIncreaseDepthsBatch: _PublicIncreaseDepthsBatchV3Api_pb2.PublicIncreaseDepthsBatchV3Api
    publicAggreDepths: _PublicAggreDepthsV3Api_pb2.PublicAggreDepthsV3Api
    publicAggreDeals: _PublicAggreDealsV3Api_pb2.PublicAggreDealsV3Api
    publicAggreBookTicker: _PublicAggreBookTickerV3Api_pb2.PublicAggreBookTickerV3Api
    symbol: str
    symbolId: str
    createTime: int
    sendTime: int
    def __init__(self, channel: _Optional[str] = ..., publicDeals: _Optional[_Union[_PublicDealsV3Api_pb2.PublicDealsV3Api, _Mapping]] = ..., publicIncreaseDepths: _Optional[_Union[_PublicIncreaseDepthsV3Api_pb2.PublicIncreaseDepthsV3Api, _Mapping]] = ..., publicLimitDepths: _Optional[_Union[_PublicLimitDepthsV3Api_pb2.PublicLimitDepthsV3Api, _Mapping]] = ..., privateOrders: _Optional[_Union[_PrivateOrdersV3Api_pb2.PrivateOrdersV3Api, _Mapping]] = ..., publicBookTicker: _Optional[_Union[_PublicBookTickerV3Api_pb2.PublicBookTickerV3Api, _Mapping]] = ..., privateDeals: _Optional[_Union[_PrivateDealsV3Api_pb2.PrivateDealsV3Api, _Mapping]] = ..., privateAccount: _Optional[_Union[_PrivateAccountV3Api_pb2.PrivateAccountV3Api, _Mapping]] = ..., publicSpotKline: _Optional[_Union[_PublicSpotKlineV3Api_pb2.PublicSpotKlineV3Api, _Mapping]] = ..., publicMiniTicker: _Optional[_Union[_PublicMiniTickerV3Api_pb2.PublicMiniTickerV3Api, _Mapping]] = ..., publicMiniTickers: _Optional[_Union[_PublicMiniTickersV3Api_pb2.PublicMiniTickersV3Api, _Mapping]] = ..., publicBookTickerBatch: _Optional[_Union[_PublicBookTickerBatchV3Api_pb2.PublicBookTickerBatchV3Api, _Mapping]] = ..., publicIncreaseDepthsBatch: _Optional[_Union[_PublicIncreaseDepthsBatchV3Api_pb2.PublicIncreaseDepthsBatchV3Api, _Mapping]] = ..., publicAggreDepths: _Optional[_Union[_PublicAggreDepthsV3Api_pb2.PublicAggreDepthsV3Api, _Mapping]] = ..., publicAggreDeals: _Optional[_Union[_PublicAggreDealsV3Api_pb2.PublicAggreDealsV3Api, _Mapping]] = ..., publicAggreBookTicker: _Optional[_Union[_PublicAggreBookTickerV3Api_pb2.PublicAggreBookTickerV3Api, _Mapping]] = ..., symbol: _Optional[str] = ..., symbolId: _Optional[str] = ..., createTime: _Optional[int] = ..., sendTime: _Optional[int] = ...) -> None: ...



from decimal import Decimal
from typing import Optional

from hummingbot.client.config.config_validators import (
    validate_bool,
    validate_decimal,
    validate_derivative,
    validate_exchange,
    validate_int,
    validate_market_trading_pair,
)
from hummingbot.client.config.config_var import ConfigVar
from hummingbot.client.settings import AllConnectorSettings, required_exchanges


def maker_trading_pair_prompt():
    derivative = perpetual_market_making_config_map.get("derivative").value
    example = AllConnectorSettings.get_example_pairs().get(derivative)
    return "Enter the token trading pair you would like to trade on %s%s >>> " \
           % (derivative, f" (e.g. {example})" if example else "")


# strategy specific validators
def validate_derivative_trading_pair(value: str) -> Optional[str]:
    derivative = perpetual_market_making_config_map.get("derivative").value
    return validate_market_trading_pair(derivative, value)


def validate_derivative_position_mode(value: str) -> Optional[str]:
    if value not in ["One-way", "Hedge"]:
        return "Position mode can either be One-way or Hedge mode"


def order_amount_prompt() -> str:
    trading_pair = perpetual_market_making_config_map["market"].value
    base_asset, quote_asset = trading_pair.split("-")
    return f"What is the amount of {base_asset} per order? >>> "


def validate_price_source(value: str) -> Optional[str]:
    if value not in {"current_market", "external_market", "custom_api"}:
        return "Invalid price source type."


def on_validate_price_source(value: str):
    if value != "external_market":
        perpetual_market_making_config_map["price_source_derivative"].value = None
        perpetual_market_making_config_map["price_source_market"].value = None
    if value != "custom_api":
        perpetual_market_making_config_map["price_source_custom_api"].value = None
    else:
        perpetual_market_making_config_map["price_type"].value = "custom"


def validate_price_type(value: str) -> Optional[str]:
    error = None
    price_source = perpetual_market_making_config_map.get("price_source").value
    if price_source != "custom_api":
        valid_values = {"mid_price",
                        "last_price",
                        "last_own_trade_price",
                        "best_bid",
                        "best_ask"}
        if value not in valid_values:
            error = "Invalid price type."
    elif value != "custom":
        error = "Invalid price type."
    return error


def price_source_market_prompt() -> str:
    external_market = perpetual_market_making_config_map.get("price_source_derivative").value
    return f'Enter the token trading pair on {external_market} >>> '


def validate_price_source_derivative(value: str) -> Optional[str]:
    if value == perpetual_market_making_config_map.get("derivative").value:
        return "Price source derivative cannot be the same as maker derivative."
    if validate_derivative(value) is not None and validate_exchange(value) is not None:
        return "Price source must must be a valid exchange or derivative connector."


def on_validated_price_source_derivative(value: str):
    if value is None:
        perpetual_market_making_config_map["price_source_market"].value = None


def validate_price_source_market(value: str) -> Optional[str]:
    market = perpetual_market_making_config_map.get("price_source_derivative").value
    return validate_market_trading_pair(market, value)


def validate_price_floor_ceiling(value: str) -> Optional[str]:
    try:
        decimal_value = Decimal(value)
    except Exception:
        return f"{value} is not in decimal format."
    if not (decimal_value == Decimal("-1") or decimal_value > Decimal("0")):
        return "Value must be more than 0 or -1 to disable this feature."


def derivative_on_validated(value: str):
    required_exchanges.add(value)


perpetual_market_making_config_map = {
    "strategy":
        ConfigVar(key="strategy",
                  prompt=None,
                  default="perpetual_market_making"),
    "derivative":
        ConfigVar(key="derivative",
                  prompt="Enter your maker derivative connector exchange name >>> ",
                  validator=validate_derivative,
                  on_validated=derivative_on_validated,
                  prompt_on_new=True),
    "market":
        ConfigVar(key="market",
                  prompt=maker_trading_pair_prompt,
                  validator=validate_derivative_trading_pair,
                  prompt_on_new=True),
    "leverage":
        ConfigVar(key="leverage",
                  prompt="How much leverage do you want to use? "
                         "(Binance Perpetual supports up to 75X for most pairs) >>> ",
                  type_str="int",
                  validator=lambda v: validate_int(v, min_value=0, inclusive=False),
                  prompt_on_new=True),
    "position_mode":
        ConfigVar(key="position_mode",
                  prompt="Which position mode do you want to use? (One-way/Hedge) >>> ",
                  validator=validate_derivative_position_mode,
                  type_str="str",
                  default="One-way",
                  prompt_on_new=True),
    "bid_spread":
        ConfigVar(key="bid_spread",
                  prompt="How far away from the mid price do you want to place the "
                         "first bid order? (Enter 1 to indicate 1%) >>> ",
                  type_str="decimal",
                  validator=lambda v: validate_decimal(v, 0, 100, inclusive=False),
                  prompt_on_new=True),
    "ask_spread":
        ConfigVar(key="ask_spread",
                  prompt="How far away from the mid price do you want to place the "
                         "first ask order? (Enter 1 to indicate 1%) >>> ",
                  type_str="decimal",
                  validator=lambda v: validate_decimal(v, 0, 100, inclusive=False),
                  prompt_on_new=True),
    "minimum_spread":
        ConfigVar(key="minimum_spread",
                  prompt="At what minimum spread should the bot automatically cancel orders? (Enter 1 for 1%) >>> ",
                  required_if=lambda: False,
                  type_str="decimal",
                  default=Decimal(-100),
                  validator=lambda v: validate_decimal(v, -100, 100, True)),
    "order_refresh_time":
        ConfigVar(key="order_refresh_time",
                  prompt="How often do you want to cancel and replace bids and asks "
                         "(in seconds)? >>> ",
                  type_str="float",
                  validator=lambda v: validate_decimal(v, 0, inclusive=False),
                  prompt_on_new=True),
    "order_refresh_tolerance_pct":
        ConfigVar(key="order_refresh_tolerance_pct",
                  prompt="Enter the percent change in price needed to refresh orders at each cycle "
                         "(Enter 1 to indicate 1%) >>> ",
                  type_str="decimal",
                  default=Decimal("0"),
                  validator=lambda v: validate_decimal(v, -10, 10, inclusive=True)),
    "order_amount":
        ConfigVar(key="order_amount",
                  prompt=order_amount_prompt,
                  type_str="decimal",
                  validator=lambda v: validate_decimal(v, min_value=Decimal("0"), inclusive=False),
                  prompt_on_new=True),
    "long_profit_taking_spread":
        ConfigVar(key="long_profit_taking_spread",
                  prompt="At what spread from the entry price do you want to place a short order to reduce position? (Enter 1 for 1%) >>> ",
                  type_str="decimal",
                  default=Decimal("0"),
                  validator=lambda v: validate_decimal(v, 0, 100, True),
                  prompt_on_new=True),
    "short_profit_taking_spread":
        ConfigVar(key="short_profit_taking_spread",
                  prompt="At what spread from the position entry price do you want to place a long order to reduce position? (Enter 1 for 1%) >>> ",
                  type_str="decimal",
                  default=Decimal("0"),
                  validator=lambda v: validate_decimal(v, 0, 100, True),
                  prompt_on_new=True),
    "stop_loss_spread":
        ConfigVar(key="stop_loss_spread",
                  prompt="At what spread from position entry price do you want to place stop_loss order? (Enter 1 for 1%) >>> ",
                  type_str="decimal",
                  default=Decimal("0"),
                  validator=lambda v: validate_decimal(v, 0, 101, False),
                  prompt_on_new=True),
    "time_between_stop_loss_orders":
        ConfigVar(key="time_between_stop_loss_orders",
                  prompt="How much time should pass before refreshing a stop loss order that has not been executed? (in seconds) >>> ",
                  type_str="float",
                  default=60,
                  validator=lambda v: validate_decimal(v, 0, inclusive=False),
                  prompt_on_new=True),
    "stop_loss_slippage_buffer":
        ConfigVar(key="stop_loss_slippage_buffer",
                  prompt="How much buffer should be added in stop loss orders' price to account for slippage? (Enter 1 for 1%)? >>> ",
                  type_str="decimal",
                  default=Decimal("0.5"),
                  validator=lambda v: validate_decimal(v, 0, inclusive=True),
                  prompt_on_new=True),
    "price_ceiling":
        ConfigVar(key="price_ceiling",
                  prompt="Enter the price point above which only sell orders will be placed "
                         "(Enter -1 to deactivate this feature) >>> ",
                  type_str="decimal",
                  default=Decimal("-1"),
                  validator=validate_price_floor_ceiling),
    "price_floor":
        ConfigVar(key="price_floor",
                  prompt="Enter the price below which only buy orders will be placed "
                         "(Enter -1 to deactivate this feature) >>> ",
                  type_str="decimal",
                  default=Decimal("-1"),
                  validator=validate_price_floor_ceiling),
    "order_levels":
        ConfigVar(key="order_levels",
                  prompt="How many orders do you want to place on both sides? >>> ",
                  type_str="int",
                  validator=lambda v: validate_int(v, min_value=0, inclusive=False),
                  default=1),
    "order_level_amount":
        ConfigVar(key="order_level_amount",
                  prompt="How much do you want to increase or decrease the order size for each "
                         "additional order? (decrease < 0 > increase) >>> ",
                  required_if=lambda: perpetual_market_making_config_map.get("order_levels").value > 1,
                  type_str="decimal",
                  validator=lambda v: validate_decimal(v),
                  default=0),
    "order_level_spread":
        ConfigVar(key="order_level_spread",
                  prompt="Enter the price increments (as percentage) for subsequent "
                         "orders? (Enter 1 to indicate 1%) >>> ",
                  required_if=lambda: perpetual_market_making_config_map.get("order_levels").value > 1,
                  type_str="decimal",
                  validator=lambda v: validate_decimal(v, 0, 100, inclusive=False),
                  default=Decimal("1")),
    "filled_order_delay":
        ConfigVar(key="filled_order_delay",
                  prompt="How long do you want to wait before placing the next order "
                         "if your order gets filled (in seconds)? >>> ",
                  type_str="float",
                  validator=lambda v: validate_decimal(v, min_value=0, inclusive=False),
                  default=60),
    "order_optimization_enabled":
        ConfigVar(key="order_optimization_enabled",
                  prompt="Do you want to enable best bid ask jumping? (Yes/No) >>> ",
                  type_str="bool",
                  default=False,
                  validator=validate_bool),
    "ask_order_optimization_depth":
        ConfigVar(key="ask_order_optimization_depth",
                  prompt="How deep do you want to go into the order book for calculating "
                         "the top ask, ignoring dust orders on the top "
                         "(expressed in base asset amount)? >>> ",
                  required_if=lambda: perpetual_market_making_config_map.get("order_optimization_enabled").value,
                  type_str="decimal",
                  validator=lambda v: validate_decimal(v, min_value=0),
                  default=0),
    "bid_order_optimization_depth":
        ConfigVar(key="bid_order_optimization_depth",
                  prompt="How deep do you want to go into the order book for calculating "
                         "the top bid, ignoring dust orders on the top "
                         "(expressed in base asset amount)? >>> ",
                  required_if=lambda: perpetual_market_making_config_map.get("order_optimization_enabled").value,
                  type_str="decimal",
                  validator=lambda v: validate_decimal(v, min_value=0),
                  default=0),
    "price_source":
        ConfigVar(key="price_source",
                  prompt="Which price source to use? (current_market/external_market/custom_api) >>> ",
                  type_str="str",
                  default="current_market",
                  validator=validate_price_source,
                  on_validated=on_validate_price_source),
    "price_type":
        ConfigVar(key="price_type",
                  prompt="Which price type to use? (mid_price/last_price/last_own_trade_price/best_bid/best_ask) >>> ",
                  type_str="str",
                  required_if=lambda: perpetual_market_making_config_map.get("price_source").value != "custom_api",
                  default="mid_price",
                  validator=validate_price_type),
    "price_source_derivative":
        ConfigVar(key="price_source_derivative",
                  prompt="Enter external price source connector name or derivative name >>> ",
                  required_if=lambda: perpetual_market_making_config_map.get("price_source").value == "external_market",
                  type_str="str",
                  validator=validate_price_source_derivative,
                  on_validated=on_validated_price_source_derivative),
    "price_source_market":
        ConfigVar(key="price_source_market",
                  prompt=price_source_market_prompt,
                  required_if=lambda: perpetual_market_making_config_map.get("price_source").value == "external_market",
                  type_str="str",
                  validator=validate_price_source_market),
    "price_source_custom_api":
        ConfigVar(key="price_source_custom_api",
                  prompt="Enter pricing API URL >>> ",
                  required_if=lambda: perpetual_market_making_config_map.get("price_source").value == "custom_api",
                  type_str="str"),
    "custom_api_update_interval":
        ConfigVar(key="custom_api_update_interval",
                  prompt="Enter custom API update interval in second (default: 5.0, min: 0.5) >>> ",
                  required_if=lambda: False,
                  default=float(5),
                  type_str="float",
                  validator=lambda v: validate_decimal(v, Decimal("0.5"))),
    "order_override":
        ConfigVar(key="order_override",
                  prompt=None,
                  required_if=lambda: False,
                  default=None,
                  type_str="json"),
}



import asyncio
from decimal import Decimal
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from hummingbot.connector.exchange.ascend_ex import ascend_ex_constants as CONSTANTS, ascend_ex_web_utils as web_utils
from hummingbot.core.data_type.common import TradeType
from hummingbot.core.data_type.order_book_message import OrderBookMessage, OrderBookMessageType
from hummingbot.core.data_type.order_book_tracker_data_source import OrderBookTrackerDataSource
from hummingbot.core.web_assistant.connections.data_types import RESTMethod, WSJSONRequest
from hummingbot.core.web_assistant.web_assistants_factory import WebAssistantsFactory
from hummingbot.core.web_assistant.ws_assistant import WSAssistant
from hummingbot.logger import HummingbotLogger

if TYPE_CHECKING:
    from hummingbot.connector.exchange.ascend_ex.ascend_ex_exchange import AscendExExchange


class AscendExAPIOrderBookDataSource(OrderBookTrackerDataSource):
    _logger: Optional[HummingbotLogger] = None
    _DYNAMIC_SUBSCRIBE_ID_START = 100
    _next_subscribe_id: int = _DYNAMIC_SUBSCRIBE_ID_START

    def __init__(
        self,
        trading_pairs: List[str],
        connector: "AscendExExchange",
        api_factory: Optional[WebAssistantsFactory] = None,
    ):
        super().__init__(trading_pairs)
        self._connector = connector
        self._trade_messages_queue_key = CONSTANTS.TRADE_TOPIC_ID
        self._diff_messages_queue_key = CONSTANTS.DIFF_TOPIC_ID
        self._api_factory = api_factory

    async def get_last_traded_prices(self, trading_pairs: List[str], domain: Optional[str] = None) -> Dict[str, float]:
        return await self._connector.get_last_traded_prices(trading_pairs=trading_pairs)

    async def _request_order_book_snapshot(self, trading_pair: str) -> Dict[str, Any]:
        """
        Retrieves a copy of the full order book from the exchange, for a particular trading pair.

        :param trading_pair: the trading pair for which the order book will be retrieved

        :return: the response from the exchange (JSON dictionary)
        """
        params = {"symbol": await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)}

        rest_assistant = await self._api_factory.get_rest_assistant()
        data = await rest_assistant.execute_request(
            url=web_utils.public_rest_url(path_url=CONSTANTS.DEPTH_PATH_URL),
            params=params,
            method=RESTMethod.GET,
            throttler_limit_id=CONSTANTS.DEPTH_PATH_URL,
        )

        return data

    async def _subscribe_channels(self, ws: WSAssistant):
        """
        Subscribes to the trade events and diff orders events through the provided websocket connection.
        :param ws: the websocket assistant used to connect to the exchange
        """
        try:
            for trading_pair in self._trading_pairs:
                trading_symbol = await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
                for topic in [CONSTANTS.DIFF_TOPIC_ID, CONSTANTS.TRADE_TOPIC_ID]:
                    payload = {"op": CONSTANTS.SUB_ENDPOINT_NAME, "ch": f"{topic}:{trading_symbol}"}
                    await ws.send(WSJSONRequest(payload=payload))

            self.logger().info("Subscribed to public order book and trade channels...")
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().error(
                "Unexpected error occurred subscribing to order book trading and delta streams...", exc_info=True
            )
            raise

    async def _connected_websocket_assistant(self) -> WSAssistant:
        ws: WSAssistant = await self._api_factory.get_ws_assistant()
        await ws.connect(ws_url=f"{CONSTANTS.WS_URL}/{CONSTANTS.STREAM_PATH_URL}")
        return ws

    async def _order_book_snapshot(self, trading_pair: str) -> OrderBookMessage:
        snapshot_response: Dict[str, Any] = await self._request_order_book_snapshot(trading_pair)
        snapshot_timestamp = float(snapshot_response["data"]["data"]["ts"]) / 1000

        order_book_message_content = {
            "trading_pair": trading_pair,
            "update_id": snapshot_timestamp,
            "bids": snapshot_response["data"]["data"]["bids"],
            "asks": snapshot_response["data"]["data"]["asks"],
        }
        snapshot_msg: OrderBookMessage = OrderBookMessage(
            OrderBookMessageType.SNAPSHOT, order_book_message_content, snapshot_timestamp
        )

        return snapshot_msg

    async def _parse_trade_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        trading_pair = await self._connector.trading_pair_associated_to_exchange_symbol(symbol=raw_message["symbol"])
        for trade_data in raw_message["data"]:
            timestamp: float = trade_data["ts"] / 1000
            message_content = {
                "trade_id": timestamp,  # trade id isn't provided so using timestamp instead
                "trading_pair": trading_pair,
                "trade_type": float(TradeType.BUY.value) if trade_data["bm"] else float(TradeType.SELL.value),
                "amount": Decimal(trade_data["q"]),
                "price": Decimal(trade_data["p"]),
            }
            trade_message: Optional[OrderBookMessage] = OrderBookMessage(
                message_type=OrderBookMessageType.TRADE, content=message_content, timestamp=timestamp
            )

            message_queue.put_nowait(trade_message)

    async def _parse_order_book_diff_message(self, raw_message: Dict[str, Any], message_queue: asyncio.Queue):
        diff_data: Dict[str, Any] = raw_message["data"]
        timestamp: float = diff_data["ts"] / 1000

        trading_pair = await self._connector.trading_pair_associated_to_exchange_symbol(symbol=raw_message["symbol"])

        message_content = {
            "trading_pair": trading_pair,
            "update_id": timestamp,
            "bids": diff_data["bids"],
            "asks": diff_data["asks"],
        }
        diff_message: OrderBookMessage = OrderBookMessage(OrderBookMessageType.DIFF, message_content, timestamp)

        message_queue.put_nowait(diff_message)

    def _channel_originating_message(self, event_message: Dict[str, Any]) -> str:
        channel = ""
        if "data" in event_message:
            event_channel = event_message.get("m")
            if event_channel == CONSTANTS.TRADE_TOPIC_ID:
                channel = self._trade_messages_queue_key
            if event_channel == CONSTANTS.DIFF_TOPIC_ID:
                channel = self._diff_messages_queue_key
        return channel

    async def _process_message_for_unknown_channel(
        self, event_message: Dict[str, Any], websocket_assistant: WSAssistant
    ):
        """
        Processes a message coming from a not identified channel.
        Does nothing by default but allows subclasses to reimplement

        :param event_message: the event received through the websocket connection
        :param websocket_assistant: the websocket connection to use to interact with the exchange
        """
        if event_message.get("m") == "ping":
            pong_payloads = {"op": "pong"}
            pong_request = WSJSONRequest(payload=pong_payloads)
            await websocket_assistant.send(request=pong_request)

    @classmethod
    def _get_next_subscribe_id(cls) -> int:
        subscribe_id = cls._next_subscribe_id
        cls._next_subscribe_id += 1
        return subscribe_id

    async def subscribe_to_trading_pair(self, trading_pair: str) -> bool:
        """
        Subscribe to order book and trade channels for a single trading pair.

        :param trading_pair: the trading pair to subscribe to
        :return: True if successful, False otherwise
        """
        if self._ws_assistant is None:
            self.logger().warning("Cannot subscribe: WebSocket connection not established")
            return False

        try:
            trading_symbol = await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
            for topic in [CONSTANTS.DIFF_TOPIC_ID, CONSTANTS.TRADE_TOPIC_ID]:
                payload = {"op": CONSTANTS.SUB_ENDPOINT_NAME, "ch": f"{topic}:{trading_symbol}"}
                await self._ws_assistant.send(WSJSONRequest(payload=payload))

            self.add_trading_pair(trading_pair)
            self.logger().info(f"Subscribed to public order book and trade channels of {trading_pair}...")
            return True
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().error(
                f"Unexpected error occurred subscribing to {trading_pair}...",
                exc_info=True
            )
            return False

    async def unsubscribe_from_trading_pair(self, trading_pair: str) -> bool:
        """
        Unsubscribe from order book and trade channels for a single trading pair.

        :param trading_pair: the trading pair to unsubscribe from
        :return: True if successful, False otherwise
        """
        if self._ws_assistant is None:
            self.logger().warning("Cannot unsubscribe: WebSocket connection not established")
            return False

        try:
            trading_symbol = await self._connector.exchange_symbol_associated_to_pair(trading_pair=trading_pair)
            for topic in [CONSTANTS.DIFF_TOPIC_ID, CONSTANTS.TRADE_TOPIC_ID]:
                payload = {"op": "unsub", "ch": f"{topic}:{trading_symbol}"}
                await self._ws_assistant.send(WSJSONRequest(payload=payload))

            self.remove_trading_pair(trading_pair)
            self.logger().info(f"Unsubscribed from public order book and trade channels of {trading_pair}...")
            return True
        except asyncio.CancelledError:
            raise
        except Exception:
            self.logger().error(
                f"Unexpected error occurred unsubscribing from {trading_pair}...",
                exc_info=True
            )
            return False



# distutils: sources=['hummingbot/core/cpp/Utils.cpp', 'hummingbot/core/cpp/LimitOrder.cpp', 'hummingbot/core/cpp/OrderExpirationEntry.cpp']

import asyncio
import math
import random
from collections import defaultdict, deque
from decimal import Decimal, ROUND_DOWN
from typing import Callable, Dict, List, Optional, Tuple, TYPE_CHECKING

from cpython cimport PyObject
from cython.operator cimport address, dereference as deref, postincrement as inc
from libcpp cimport bool as cppbool
from libcpp.vector cimport vector

from hummingbot.connector.budget_checker import BudgetChecker
from hummingbot.connector.connector_metrics_collector import DummyMetricsCollector
from hummingbot.connector.exchange.paper_trade.trading_pair import TradingPair
from hummingbot.connector.exchange_base import ExchangeBase
from hummingbot.core.clock cimport Clock
from hummingbot.core.clock import Clock
from hummingbot.core.data_type.cancellation_result import CancellationResult
from hummingbot.core.data_type.common import OrderType, TradeType
from hummingbot.core.data_type.composite_order_book import CompositeOrderBook
from hummingbot.core.data_type.composite_order_book cimport CompositeOrderBook
from hummingbot.core.data_type.limit_order import LimitOrder
from hummingbot.core.data_type.limit_order cimport c_create_limit_order_from_cpp_limit_order
from hummingbot.core.data_type.order_book cimport OrderBook
from hummingbot.core.data_type.order_book_tracker import OrderBookTracker
from hummingbot.core.data_type.order_candidate import OrderCandidate
from hummingbot.core.event.event_listener cimport EventListener
from hummingbot.core.event.events import (
    BuyOrderCompletedEvent,
    BuyOrderCreatedEvent,
    MarketEvent,
    MarketOrderFailureEvent,
    OrderBookEvent,
    OrderBookTradeEvent,
    OrderCancelledEvent,
    OrderFilledEvent,
    SellOrderCompletedEvent,
    SellOrderCreatedEvent,
)
from hummingbot.core.network_iterator import NetworkStatus
from hummingbot.core.Utils cimport getIteratorFromReverseIterator, reverse_iterator
from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.core.utils.estimate_fee import build_trade_fee

if TYPE_CHECKING:
    from hummingbot.client.config.config_helpers import ClientConfigAdapter

ptm_logger = None
s_decimal_0 = Decimal(0)


cdef class QuantizationParams:
    def __init__(self,
                 str trading_pair,
                 int price_precision,
                 int price_decimals,
                 int order_size_precision,
                 int order_size_decimals):
        self.trading_pair = trading_pair
        self.price_precision = price_precision
        self.price_decimals = price_decimals
        self.order_size_precision = order_size_precision
        self.order_size_decimals = order_size_decimals

    def __repr__(self) -> str:
        return (f"QuantizationParams('{self.trading_pair}', {self.price_precision}, {self.price_decimals}, "
                f"{self.order_size_precision}, {self.order_size_decimals})")


cdef class QueuedOrder:
    cdef:
        double create_timestamp
        str _order_id
        bint _is_buy
        str _trading_pair
        object _amount

    def __init__(self, create_timestamp: float, order_id: str, is_buy: bool, trading_pair: str, amount: Decimal):
        self.create_timestamp = create_timestamp
        self._order_id = order_id
        self._is_buy = is_buy
        self._trading_pair = trading_pair
        self._amount = amount

    @property
    def timestamp(self) -> double:
        return self.create_timestamp

    @property
    def order_id(self) -> str:
        return self._order_id

    @property
    def is_buy(self) -> bint:
        return self._is_buy

    @property
    def trading_pair(self) -> str:
        return self._trading_pair

    @property
    def amount(self) -> Decimal:
        return self._amount

    def __repr__(self) -> str:
        return (f"QueuedOrder({self.create_timestamp}, '{self.order_id}', {self.is_buy}, '{self.trading_pair}', "
                f"{self.amount})")


cdef class OrderBookTradeListener(EventListener):
    cdef:
        ExchangeBase _market

    def __init__(self, market: ExchangeBase):
        super().__init__()
        self._market = market

    cdef c_call(self, object event_object):
        try:
            self._market.match_trade_to_limit_orders(event_object)
        except Exception as e:
            self.logger().error("Error call trade listener.", exc_info=True)

cdef class OrderBookMarketOrderFillListener(EventListener):
    cdef:
        ExchangeBase _market

    def __init__(self, market: ExchangeBase):
        super().__init__()
        self._market = market

    cdef c_call(self, object event_object):

        if event_object.trading_pair not in self._market.order_books or event_object.order_type != OrderType.MARKET:
            return
        order_book = self._market.order_books[event_object.trading_pair]
        order_book.record_filled_order(event_object)


cdef class PaperTradeExchange(ExchangeBase):
    TRADE_EXECUTION_DELAY = 5.0
    ORDER_FILLED_EVENT_TAG = MarketEvent.OrderFilled.value
    SELL_ORDER_COMPLETED_EVENT_TAG = MarketEvent.SellOrderCompleted.value
    BUY_ORDER_COMPLETED_EVENT_TAG = MarketEvent.BuyOrderCompleted.value
    MARKET_ORDER_CANCELED_EVENT_TAG = MarketEvent.OrderCancelled.value
    MARKET_ORDER_FAILURE_EVENT_TAG = MarketEvent.OrderFailure.value
    ORDER_BOOK_TRADE_EVENT_TAG = OrderBookEvent.TradeEvent.value
    MARKET_SELL_ORDER_CREATED_EVENT_TAG = MarketEvent.SellOrderCreated.value
    MARKET_BUY_ORDER_CREATED_EVENT_TAG = MarketEvent.BuyOrderCreated.value

    def __init__(
        self,
        order_book_tracker: OrderBookTracker,
        target_market: Callable,
        exchange_name: str,
        balance_asset_limit: Optional[Dict[str, Dict[str, Decimal]]] = None,
        rate_limits_share_pct: Decimal = Decimal("100"),
    ):
        order_book_tracker.data_source.order_book_create_function = lambda: CompositeOrderBook()
        super().__init__(balance_asset_limit, rate_limits_share_pct)
        self._set_order_book_tracker(order_book_tracker)
        self._budget_checker = BudgetChecker(exchange=self)
        self._exchange_name = exchange_name
        self._account_balances = {}
        self._account_available_balances = {}
        self._paper_trade_market_initialized = False
        self._trading_pairs = {}
        self._queued_orders = deque()
        self._quantization_params = {}
        self._order_book_trade_listener = OrderBookTradeListener(self)
        self._target_market = target_market
        self._market_order_filled_listener = OrderBookMarketOrderFillListener(self)
        self.c_add_listener(self.ORDER_FILLED_EVENT_TAG, self._market_order_filled_listener)

        # Trade volume metrics should never be gather for paper trade connector
        self._trade_volume_metric_collector = DummyMetricsCollector()

    @property
    def budget_checker(self) -> BudgetChecker:
        return self._budget_checker

    @classmethod
    def random_order_id(cls, order_side: str, trading_pair: str) -> str:
        vals = [random.choice(range(0, 256)) for i in range(0, 13)]
        return f"{order_side}://" + trading_pair + "/" + "".join([f"{val:02x}" for val in vals])

    def init_paper_trade_market(self):
        for trading_pair_str, order_book in self.order_book_tracker.order_books.items():
            assert type(order_book) is CompositeOrderBook
            base_asset, quote_asset = self.split_trading_pair(trading_pair_str)
            self._trading_pairs[self._target_market.convert_from_exchange_trading_pair(trading_pair_str)] = TradingPair(trading_pair_str, base_asset, quote_asset)
            (<CompositeOrderBook>order_book).c_add_listener(
                self.ORDER_BOOK_TRADE_EVENT_TAG,
                self._order_book_trade_listener
            )

    def split_trading_pair(self, trading_pair: str) -> Tuple[str, str]:
        return self._target_market.split_trading_pair(trading_pair)

    #  <editor-fold desc="Property">
    @property
    def trading_pair(self) -> Dict[str, TradingPair]:
        return self._trading_pairs

    @property
    def trading_pairs(self) -> List[str]:
        return [trading_pair for trading_pair in self._trading_pairs]

    @property
    def name(self) -> str:
        return self._exchange_name

    @property
    def display_name(self) -> str:
        return f"{self._exchange_name}_PaperTrade"

    @property
    def order_books(self) -> Dict[str, CompositeOrderBook]:
        return self.order_book_tracker.order_books

    @property
    def status_dict(self) -> Dict[str, bool]:
        return {
            "order_books_initialized": self.order_book_tracker and len(self.order_book_tracker.order_books) > 0
        }

    @property
    def ready(self):
        if not self.order_book_tracker.ready:
            return False
        if all(self.status_dict.values()):
            if not self._paper_trade_market_initialized:
                self.init_paper_trade_market()
                self._paper_trade_market_initialized = True
            return True
        else:
            return False

    @property
    def queued_orders(self) -> List[QueuedOrder]:
        return self._queued_orders

    @property
    def limit_orders(self) -> List[LimitOrder]:
        cdef:
            LimitOrdersIterator map_it
            SingleTradingPairLimitOrders *single_trading_pair_collection_ptr
            SingleTradingPairLimitOrdersIterator collection_it
            SingleTradingPairLimitOrdersRIterator collection_rit
            const CPPLimitOrder *cpp_limit_order_ptr
            list retval = []

        map_it = self._bid_limit_orders.begin()
        while map_it != self._bid_limit_orders.end():
            single_trading_pair_collection_ptr = address(deref(map_it).second)
            collection_rit = single_trading_pair_collection_ptr.rbegin()
            while collection_rit != single_trading_pair_collection_ptr.rend():
                cpp_limit_order_ptr = address(deref(collection_rit))
                retval.append(c_create_limit_order_from_cpp_limit_order(deref(cpp_limit_order_ptr)))
                inc(collection_rit)
            inc(map_it)

        map_it = self._ask_limit_orders.begin()
        while map_it != self._ask_limit_orders.end():
            single_trading_pair_collection_ptr = address(deref(map_it).second)
            collection_it = single_trading_pair_collection_ptr.begin()
            while collection_it != single_trading_pair_collection_ptr.end():
                cpp_limit_order_ptr = address(deref(collection_it))
                retval.append(c_create_limit_order_from_cpp_limit_order(deref(cpp_limit_order_ptr)))
                inc(collection_it)
            inc(map_it)

        return retval

    @property
    def on_hold_balances(self) -> Dict[str, Decimal]:
        _on_hold_balances = defaultdict(Decimal)
        for limit_order in self.limit_orders:
            if limit_order.is_buy:
                _on_hold_balances[limit_order.quote_currency] += limit_order.quantity * limit_order.price
            else:
                _on_hold_balances[limit_order.base_currency] += limit_order.quantity
        return _on_hold_balances

    @property
    def available_balances(self) -> Dict[str, Decimal]:
        _available_balances = self._account_balances.copy()
        for trading_pair_str, balance in _available_balances.items():
            _available_balances[trading_pair_str] -= self.on_hold_balances[trading_pair_str]
        return _available_balances

    # </editor-fold>

    cdef c_start(self, Clock clock, double timestamp):
        ExchangeBase.c_start(self, clock, timestamp)

    async def start_network(self):
        await self.stop_network()
        self.order_book_tracker.start()

    async def stop_network(self):
        self.order_book_tracker.stop()

    async def check_network(self) -> NetworkStatus:
        return NetworkStatus.CONNECTED

    cdef c_set_balance(self, str currency, object balance):
        self._account_balances[currency.upper()] = Decimal(balance)

    cdef object c_get_balance(self, str currency):
        if currency.upper() not in self._account_balances:
            self.logger().warning(f"Account balance does not have asset {currency.upper()}.")
            return Decimal(0.0)
        return self._account_balances[currency.upper()]

    cdef c_tick(self, double timestamp):
        ExchangeBase.c_tick(self, timestamp)
        self.c_process_market_orders()
        self.c_process_crossed_limit_orders()

    cdef str c_buy(self,
                   str trading_pair_str,
                   object amount,
                   object order_type=OrderType.MARKET,
                   object price=s_decimal_0,
                   dict kwargs={}):
        if trading_pair_str not in self._trading_pairs:
            raise ValueError(f"Trading pair '{trading_pair_str}' does not existing in current data set.")

        cdef:
            str order_id = self.random_order_id("buy", trading_pair_str)
            str quote_asset = self._trading_pairs[trading_pair_str].quote_asset
            string cpp_order_id = order_id.encode("utf8")
            string cpp_trading_pair_str = trading_pair_str.encode("utf8")
            string cpp_base_asset = self._trading_pairs[trading_pair_str].base_asset.encode("utf8")
            string cpp_quote_asset = quote_asset.encode("utf8")
            string cpp_position = "NIL".encode("utf8")
            LimitOrdersIterator map_it
            SingleTradingPairLimitOrders *limit_orders_collection_ptr = NULL
            pair[LimitOrders.iterator, cppbool] insert_result

        quantized_price = (self.c_quantize_order_price(trading_pair_str, price)
                           if order_type is OrderType.LIMIT
                           else s_decimal_0)
        quantized_amount = self.c_quantize_order_amount(trading_pair_str, amount)
        if order_type is OrderType.MARKET:
            self._queued_orders.append(QueuedOrder(self._current_timestamp, order_id, True, trading_pair_str,
                                                   quantized_amount))
        elif order_type is OrderType.LIMIT:

            map_it = self._bid_limit_orders.find(cpp_trading_pair_str)

            if map_it == self._bid_limit_orders.end():
                insert_result = self._bid_limit_orders.insert(LimitOrdersPair(cpp_trading_pair_str,
                                                                              SingleTradingPairLimitOrders()))
                map_it = insert_result.first
            limit_orders_collection_ptr = address(deref(map_it).second)
            limit_orders_collection_ptr.insert(CPPLimitOrder(
                cpp_order_id,
                cpp_trading_pair_str,
                True,
                cpp_base_asset,
                cpp_quote_asset,
                <PyObject *> quantized_price,
                <PyObject *> quantized_amount,
                <PyObject *> None,
                int(self._current_timestamp * 1e6),
                0,
                cpp_position,
            ))
        safe_ensure_future(self.trigger_event_async(
            self.MARKET_BUY_ORDER_CREATED_EVENT_TAG,
            BuyOrderCreatedEvent(self._current_timestamp,
                                 order_type,
                                 trading_pair_str,
                                 quantized_amount,
                                 quantized_price,
                                 order_id,
                                 self._current_timestamp)))
        return order_id

    cdef str c_sell(self,
                    str trading_pair_str,
                    object amount,
                    object order_type=OrderType.MARKET,
                    object price=s_decimal_0,
                    dict kwargs={}):

        if trading_pair_str not in self._trading_pairs:
            raise ValueError(f"Trading pair '{trading_pair_str}' does not existing in current data set.")
        cdef:
            str order_id = self.random_order_id("sell", trading_pair_str)
            str base_asset = self._trading_pairs[trading_pair_str].base_asset
            string cpp_order_id = order_id.encode("utf8")
            string cpp_trading_pair_str = trading_pair_str.encode("utf8")
            string cpp_base_asset = base_asset.encode("utf8")
            string cpp_quote_asset = self._trading_pairs[trading_pair_str].quote_asset.encode("utf8")
            string cpp_position = "NIL".encode("utf8")
            LimitOrdersIterator map_it
            SingleTradingPairLimitOrders *limit_orders_collection_ptr = NULL
            pair[LimitOrders.iterator, cppbool] insert_result

        quantized_price = (self.c_quantize_order_price(trading_pair_str, price)
                           if order_type is OrderType.LIMIT
                           else s_decimal_0)
        quantized_amount = self.c_quantize_order_amount(trading_pair_str, amount)
        if order_type is OrderType.MARKET:
            self._queued_orders.append(QueuedOrder(self._current_timestamp, order_id, False, trading_pair_str,
                                                   quantized_amount))
        elif order_type is OrderType.LIMIT:
            map_it = self._ask_limit_orders.find(cpp_trading_pair_str)

            if map_it == self._ask_limit_orders.end():
                insert_result = self._ask_limit_orders.insert(LimitOrdersPair(cpp_trading_pair_str,
                                                                              SingleTradingPairLimitOrders()))
                map_it = insert_result.first
            limit_orders_collection_ptr = address(deref(map_it).second)
            limit_orders_collection_ptr.insert(CPPLimitOrder(
                cpp_order_id,
                cpp_trading_pair_str,
                False,
                cpp_base_asset,
                cpp_quote_asset,
                <PyObject *> quantized_price,
                <PyObject *> quantized_amount,
                <PyObject *> None,
                int(self._current_timestamp * 1e6),
                0,
                cpp_position,
            ))
        safe_ensure_future(self.trigger_event_async(
            self.MARKET_SELL_ORDER_CREATED_EVENT_TAG,
            SellOrderCreatedEvent(self._current_timestamp,
                                  order_type,
                                  trading_pair_str,
                                  quantized_amount,
                                  quantized_price,
                                  order_id,
                                  self._current_timestamp)))
        return order_id

    cdef c_execute_buy(self, str order_id, str trading_pair_str, object amount):
        cdef:
            str quote_asset = self._trading_pairs[trading_pair_str].quote_asset
            str base_asset = self._trading_pairs[trading_pair_str].base_asset
            object quote_balance = self.c_get_balance(quote_asset)
            object base_balance = self.c_get_balance(base_asset)

        order_book = self.order_books[trading_pair_str]

        buy_entries = order_book.simulate_buy(float(amount))

        # Get the weighted average price of the trade
        avg_price = Decimal(0)
        for entry in buy_entries:
            avg_price += Decimal(entry.price) * Decimal(entry.amount)
        avg_price = avg_price / amount

        order_candidate = OrderCandidate(
            trading_pair=trading_pair_str,
            # Market orders are not maker orders
            is_maker=False,
            order_type=OrderType.MARKET,
            order_side=TradeType.BUY,
            amount=amount,
            price=avg_price,
            from_total_balances=True
        )

        adjusted_order_candidate = self._budget_checker.populate_collateral_entries(order_candidate)

        # Quote currency used, including fees.
        paid_amount = adjusted_order_candidate.order_collateral.amount
        # Base currency acquired, including fees.
        acquired_amount = adjusted_order_candidate.potential_returns.amount

        # It's not possible to fulfill the order, the possible acquired amount is less than requested
        if paid_amount > quote_balance:
            self.logger().warning(f"Insufficient {quote_asset} balance available for buy order. "
                                  f"{quote_balance} {quote_asset} available vs. "
                                  f"{paid_amount} {quote_asset} required for the order.")
            self.c_trigger_event(
                self.MARKET_ORDER_FAILURE_EVENT_TAG,
                MarketOrderFailureEvent(self._current_timestamp, order_id, OrderType.MARKET)
            )
            return

        # The order was successfully executed
        self.c_set_balance(quote_asset,
                           quote_balance - paid_amount)
        self.c_set_balance(base_asset,
                           base_balance + acquired_amount)

        # add fee
        fees = build_trade_fee(
            exchange=self.name,
            is_maker=False,
            base_currency="",
            quote_currency="",
            order_type=OrderType.LIMIT,
            order_side=TradeType.BUY,
            amount=Decimal("0"),
            price=Decimal("0"),
        )

        order_filled_events = OrderFilledEvent.order_filled_events_from_order_book_rows(
            self._current_timestamp, order_id, trading_pair_str, TradeType.BUY, OrderType.MARKET,
            fees, buy_entries
        )

        for order_filled_event in order_filled_events:
            self.c_trigger_event(self.ORDER_FILLED_EVENT_TAG, order_filled_event)

        self.c_trigger_event(
            self.BUY_ORDER_COMPLETED_EVENT_TAG,
            BuyOrderCompletedEvent(self._current_timestamp,
                                   order_id,
                                   base_asset,
                                   quote_asset,
                                   acquired_amount,
                                   paid_amount,
                                   OrderType.MARKET))

    cdef c_execute_sell(self, str order_id, str trading_pair_str, object amount):
        cdef:
            str quote_asset = self._trading_pairs[trading_pair_str].quote_asset
            str base_asset = self._trading_pairs[trading_pair_str].base_asset
            object quote_balance = self.c_get_balance(quote_asset)
            object base_balance = self.c_get_balance(base_asset)

        order_book = self.order_books[trading_pair_str]

        sell_entries = order_book.simulate_sell(float(amount))

        # Get the weighted average price of the trade
        avg_price = Decimal(0)
        for entry in sell_entries:
            avg_price += Decimal(entry.price) * Decimal(entry.amount)
        avg_price = avg_price / amount

        order_candidate = OrderCandidate(
            trading_pair=trading_pair_str,
            # Market orders are not maker orders
            is_maker=False,
            order_type=OrderType.MARKET,
            order_side=TradeType.SELL,
            amount=amount,
            price=avg_price,
            from_total_balances=True
        )

        adjusted_order_candidate = self._budget_checker.populate_collateral_entries(order_candidate)

        # Base currency used, including fees.
        sold_amount = adjusted_order_candidate.order_collateral.amount
        # Quote currency acquired, including fees.
        acquired_amount = adjusted_order_candidate.potential_returns.amount

        # It's not possible to fulfill the order, the possible sold amount is less than requested
        if sold_amount > base_balance:
            self.logger().warning(f"Insufficient {base_asset} balance available for sell order. "
                                  f"{base_balance} {base_asset} available vs. "
                                  f"{amount} {base_asset} required for the order.")
            self.c_trigger_event(
                self.MARKET_ORDER_FAILURE_EVENT_TAG,
                MarketOrderFailureEvent(self._current_timestamp, order_id, OrderType.MARKET)
            )
            return

        # The order was successfully executed
        self.c_set_balance(quote_asset,
                           quote_balance + acquired_amount)
        self.c_set_balance(base_asset,
                           base_balance - sold_amount)

        # add fee
        fees = build_trade_fee(
            exchange=self.name,
            is_maker=False,
            base_currency="",
            quote_currency="",
            order_type=OrderType.LIMIT,
            order_side=TradeType.BUY,
            amount=Decimal("0"),
            price=Decimal("0"),
        )

        order_filled_events = OrderFilledEvent.order_filled_events_from_order_book_rows(
            self._current_timestamp, order_id, trading_pair_str, TradeType.SELL,
            OrderType.MARKET, fees, sell_entries
        )

        for order_filled_event in order_filled_events:
            self.c_trigger_event(self.ORDER_FILLED_EVENT_TAG, order_filled_event)

        self.c_trigger_event(
            self.SELL_ORDER_COMPLETED_EVENT_TAG,
            SellOrderCompletedEvent(self._current_timestamp,
                                    order_id,
                                    base_asset,
                                    quote_asset,
                                    sold_amount,
                                    acquired_amount,
                                    OrderType.MARKET))

    cdef c_process_market_orders(self):
        cdef:
            QueuedOrder front_order = None
        while len(self._queued_orders) > 0:
            front_order = self._queued_orders[0]
            if front_order.create_timestamp <= self._current_timestamp - self.TRADE_EXECUTION_DELAY:
                self._queued_orders.popleft()
                try:
                    if front_order.is_buy:
                        self.c_execute_buy(front_order.order_id, front_order.trading_pair, front_order.amount)
                    else:
                        self.c_execute_sell(front_order.order_id, front_order.trading_pair, front_order.amount)
                except Exception as e:
                    self.logger().error("Error executing queued order.", exc_info=True)
            else:
                return

    cdef c_delete_limit_order(self,
                              LimitOrders *limit_orders_map_ptr,
                              LimitOrdersIterator *map_it_ptr,
                              const SingleTradingPairLimitOrdersIterator orders_it):
        cdef:
            SingleTradingPairLimitOrders *orders_collection_ptr = address(deref(deref(map_it_ptr)).second)
        try:
            orders_collection_ptr.erase(orders_it)
            if orders_collection_ptr.empty():
                map_it_ptr[0] = limit_orders_map_ptr.erase(deref(map_it_ptr))
            return True
        except Exception as err:
            self.logger().error("Error deleting limit order.", exc_info=True)
            return False

    cdef c_process_limit_bid_order(self,
                                   LimitOrders *limit_orders_map_ptr,
                                   LimitOrdersIterator *map_it_ptr,
                                   SingleTradingPairLimitOrdersIterator orders_it):
        cdef:
            const CPPLimitOrder *cpp_limit_order_ptr = address(deref(orders_it))
            str trading_pair_str = cpp_limit_order_ptr.getTradingPair().decode("utf8")
            str quote_asset = cpp_limit_order_ptr.getQuoteCurrency().decode("utf8")
            str base_asset = cpp_limit_order_ptr.getBaseCurrency().decode("utf8")
            str order_id = cpp_limit_order_ptr.getClientOrderID().decode("utf8")
            object amount = <object> cpp_limit_order_ptr.getQuantity()
            object price = <object> cpp_limit_order_ptr.getPrice()
            object quote_balance = self.c_get_balance(quote_asset)
            object base_balance = self.c_get_balance(base_asset)

        order_candidate = OrderCandidate(
            trading_pair=trading_pair_str,
            is_maker=True,
            order_type=OrderType.LIMIT,
            order_side=TradeType.BUY,
            amount=amount,
            price=price,
            from_total_balances=True
        )

        adjusted_order_candidate = self._budget_checker.populate_collateral_entries(order_candidate)

        # Quote currency used, including fees.
        paid_amount = adjusted_order_candidate.order_collateral.amount
        # Base currency acquired, including fees.
        acquired_amount = adjusted_order_candidate.potential_returns.amount

        # It's not possible to fulfill the order, the possible acquired amount is less than requested
        if paid_amount > quote_balance:
            self.logger().warning(f"Not enough {quote_asset} balance to fill limit buy order on {trading_pair_str}. "
                                  f"{paid_amount:.8g} {quote_asset} needed vs. "
                                  f"{quote_balance:.8g} {quote_asset} available.")

            self.c_delete_limit_order(limit_orders_map_ptr, map_it_ptr, orders_it)
            self.c_trigger_event(self.MARKET_ORDER_CANCELED_EVENT_TAG,
                                 OrderCancelledEvent(self._current_timestamp,
                                                     order_id)
                                 )
            return

        # The order was successfully executed
        self.c_set_balance(quote_asset,
                           quote_balance - paid_amount)
        self.c_set_balance(base_asset,
                           base_balance + acquired_amount)

        # add fee
        fees = build_trade_fee(
            exchange=self.name,
            is_maker=True,
            base_currency="",
            quote_currency="",
            order_type=OrderType.LIMIT,
            order_side=TradeType.BUY,
            amount=Decimal("0"),
            price=Decimal("0"),
        )

        # Emit the trade and order completed events.
        self.c_trigger_event(
            self.ORDER_FILLED_EVENT_TAG,
            OrderFilledEvent(
                self._current_timestamp,
                order_id,
                trading_pair_str,
                TradeType.BUY,
                OrderType.LIMIT,
                <object> cpp_limit_order_ptr.getPrice(),
                <object> cpp_limit_order_ptr.getQuantity(),
                fees,
                exchange_trade_id=str(int(self._time() * 1e6))
            ))

        self.c_trigger_event(
            self.BUY_ORDER_COMPLETED_EVENT_TAG,
            BuyOrderCompletedEvent(
                self._current_timestamp,
                order_id,
                base_asset,
                quote_asset,
                acquired_amount,
                paid_amount,
                OrderType.LIMIT
            ))
        self.c_delete_limit_order(limit_orders_map_ptr, map_it_ptr, orders_it)

    cdef c_process_limit_ask_order(self,
                                   LimitOrders *limit_orders_map_ptr,
                                   LimitOrdersIterator *map_it_ptr,
                                   SingleTradingPairLimitOrdersIterator orders_it):
        cdef:
            const CPPLimitOrder *cpp_limit_order_ptr = address(deref(orders_it))
            str trading_pair_str = cpp_limit_order_ptr.getTradingPair().decode("utf8")
            str quote_asset = cpp_limit_order_ptr.getQuoteCurrency().decode("utf8")
            str base_asset = cpp_limit_order_ptr.getBaseCurrency().decode("utf8")
            str order_id = cpp_limit_order_ptr.getClientOrderID().decode("utf8")
            object amount = <object> cpp_limit_order_ptr.getQuantity()
            object price = <object> cpp_limit_order_ptr.getPrice()
            object quote_balance = self.c_get_balance(quote_asset)
            object base_balance = self.c_get_balance(base_asset)

        order_candidate = OrderCandidate(
            trading_pair=trading_pair_str,
            # Market orders are not maker orders
            is_maker=True,
            order_type=OrderType.LIMIT,
            order_side=TradeType.SELL,
            amount=amount,
            price=price,
            from_total_balances=True
        )

        adjusted_order_candidate = self._budget_checker.populate_collateral_entries(order_candidate)

        # Base currency used, including fees.
        sold_amount = adjusted_order_candidate.order_collateral.amount
        # Quote currency acquired, including fees.
        acquired_amount = adjusted_order_candidate.potential_returns.amount

        # It's not possible to fulfill the order, the possible sold amount is less than requested
        if sold_amount > base_balance:
            self.logger().warning(f"Not enough {base_asset} balance to fill limit sell order on {trading_pair_str}. "
                                  f"{sold_amount:.8g} {base_asset} needed vs. "
                                  f"{base_balance:.8g} {base_asset} available.")
            self.c_delete_limit_order(limit_orders_map_ptr, map_it_ptr, orders_it)
            self.c_trigger_event(self.MARKET_ORDER_CANCELED_EVENT_TAG,
                                 OrderCancelledEvent(self._current_timestamp,
                                                     order_id)
                                 )
            return

        # The order was successfully executed
        self.c_set_balance(quote_asset,
                           quote_balance + acquired_amount)
        self.c_set_balance(base_asset,
                           base_balance - sold_amount)

        # add fee
        fees = build_trade_fee(
            exchange=self.name,
            is_maker=True,
            base_currency="",
            quote_currency="",
            order_type=OrderType.LIMIT,
            order_side=TradeType.SELL,
            amount=Decimal("0"),
            price=Decimal("0"),
        )

        # Emit the trade and order completed events.
        self.c_trigger_event(
            self.ORDER_FILLED_EVENT_TAG,
            OrderFilledEvent(
                self._current_timestamp,
                order_id,
                trading_pair_str,
                TradeType.SELL,
                OrderType.LIMIT,
                <object> cpp_limit_order_ptr.getPrice(),
                <object> cpp_limit_order_ptr.getQuantity(),
                fees,
                exchange_trade_id=str(int(self._time() * 1e6))
            ))

        self.c_trigger_event(
            self.SELL_ORDER_COMPLETED_EVENT_TAG,
            SellOrderCompletedEvent(
                self._current_timestamp,
                order_id,
                base_asset,
                quote_asset,
                sold_amount,
                acquired_amount,
                OrderType.LIMIT
            ))
        self.c_delete_limit_order(limit_orders_map_ptr, map_it_ptr, orders_it)

    cdef c_process_limit_order(self,
                               bint is_buy,
                               LimitOrders *limit_orders_map_ptr,
                               LimitOrdersIterator *map_it_ptr,
                               SingleTradingPairLimitOrdersIterator orders_it):
        try:
            if is_buy:
                self.c_process_limit_bid_order(limit_orders_map_ptr, map_it_ptr, orders_it)
            else:
                self.c_process_limit_ask_order(limit_orders_map_ptr, map_it_ptr, orders_it)
        except Exception as e:
            self.logger().error(f"Error processing limit order.", exc_info=True)

    cdef c_process_crossed_limit_orders_for_trading_pair(self,
                                                         bint is_buy,
                                                         LimitOrders *limit_orders_map_ptr,
                                                         LimitOrdersIterator *map_it_ptr):
        """
        Trigger limit orders when the opposite side of the order book has crossed the limit order's price.
        This implies someone was ready to fill the limit order, if that limit order was on the market.

        :param is_buy: are the limit orders on the bid side?
        :param limit_orders_map_ptr: pointer to the limit orders map
        :param map_it_ptr: limit orders map iterator, which implies the trading pair being processed
        """
        cdef:
            str trading_pair = deref(deref(map_it_ptr)).first.decode("utf8")
            object opposite_order_book_price = self.c_get_price(trading_pair, is_buy)
            SingleTradingPairLimitOrders *orders_collection_ptr = address(deref(deref(map_it_ptr)).second)
            SingleTradingPairLimitOrdersIterator orders_it = orders_collection_ptr.begin()
            SingleTradingPairLimitOrdersRIterator orders_rit = orders_collection_ptr.rbegin()
            vector[SingleTradingPairLimitOrdersIterator] process_order_its
            const CPPLimitOrder *cpp_limit_order_ptr = NULL

        if is_buy:
            while orders_rit != orders_collection_ptr.rend():
                cpp_limit_order_ptr = address(deref(orders_rit))
                if opposite_order_book_price > <object>cpp_limit_order_ptr.getPrice():
                    break
                process_order_its.push_back(getIteratorFromReverseIterator(
                    <reverse_iterator[SingleTradingPairLimitOrdersIterator]>orders_rit))
                inc(orders_rit)
        else:
            while orders_it != orders_collection_ptr.end():
                cpp_limit_order_ptr = address(deref(orders_it))
                if opposite_order_book_price < <object>cpp_limit_order_ptr.getPrice():
                    break
                process_order_its.push_back(orders_it)
                inc(orders_it)

        for orders_it in process_order_its:
            self.c_process_limit_order(is_buy, limit_orders_map_ptr, map_it_ptr, orders_it)

    cdef c_process_crossed_limit_orders(self):
        cdef:
            LimitOrders *limit_orders_ptr = address(self._bid_limit_orders)
            LimitOrdersIterator map_it = limit_orders_ptr.begin()

        while map_it != limit_orders_ptr.end():
            self.c_process_crossed_limit_orders_for_trading_pair(True, limit_orders_ptr, address(map_it))
            if map_it != limit_orders_ptr.end():
                inc(map_it)

        limit_orders_ptr = address(self._ask_limit_orders)
        map_it = limit_orders_ptr.begin()

        while map_it != limit_orders_ptr.end():
            self.c_process_crossed_limit_orders_for_trading_pair(False, limit_orders_ptr, address(map_it))
            if map_it != limit_orders_ptr.end():
                inc(map_it)

    # <editor-fold desc="Event listener functions">
    cdef c_match_trade_to_limit_orders(self, object order_book_trade_event):
        """
        Trigger limit orders when incoming market orders have crossed the limit order's price.

        :param order_book_trade_event: trade event from order book
        """
        cdef:
            string cpp_trading_pair = order_book_trade_event.trading_pair.encode("utf8")
            bint is_maker_buy = order_book_trade_event.type is TradeType.SELL
            object trade_price = order_book_trade_event.price
            object trade_quantity = order_book_trade_event.amount
            LimitOrders *limit_orders_map_ptr = (address(self._bid_limit_orders)
                                                 if is_maker_buy
                                                 else address(self._ask_limit_orders))
            LimitOrdersIterator map_it = limit_orders_map_ptr.find(cpp_trading_pair)
            SingleTradingPairLimitOrders *orders_collection_ptr = NULL
            SingleTradingPairLimitOrdersIterator orders_it
            SingleTradingPairLimitOrdersRIterator orders_rit
            vector[SingleTradingPairLimitOrdersIterator] process_order_its
            const CPPLimitOrder *cpp_limit_order_ptr = NULL

        if map_it == limit_orders_map_ptr.end():
            return

        orders_collection_ptr = address(deref(map_it).second)
        if is_maker_buy:
            orders_rit = orders_collection_ptr.rbegin()
            while orders_rit != orders_collection_ptr.rend():
                cpp_limit_order_ptr = address(deref(orders_rit))
                if <object>cpp_limit_order_ptr.getPrice() <= trade_price:
                    break
                process_order_its.push_back(getIteratorFromReverseIterator(
                    <reverse_iterator[SingleTradingPairLimitOrdersIterator]>orders_rit))
                inc(orders_rit)
        else:
            orders_it = orders_collection_ptr.begin()
            while orders_it != orders_collection_ptr.end():
                cpp_limit_order_ptr = address(deref(orders_it))
                if <object>cpp_limit_order_ptr.getPrice() >= trade_price:
                    break
                process_order_its.push_back(orders_it)
                inc(orders_it)

        for orders_it in process_order_its:
            self.c_process_limit_order(is_maker_buy, limit_orders_map_ptr, address(map_it), orders_it)

    # </editor-fold>

    cdef object c_get_available_balance(self, str currency):
        return self.available_balances.get(currency.upper(), s_decimal_0)

    async def cancel_all(self, timeout_seconds: float) -> List[CancellationResult]:
        cdef:
            LimitOrders *limit_orders_map_ptr
            list cancellation_results = []
        limit_orders_map_ptr = address(self._bid_limit_orders)
        for trading_pair_str in self._trading_pairs.keys():
            results = self.c_cancel_order_from_orders_map(limit_orders_map_ptr, trading_pair_str, cancel_all=True)
            cancellation_results.extend(results)

        limit_orders_map_ptr = address(self._ask_limit_orders)
        for trading_pair_str in self._trading_pairs.keys():
            results = self.c_cancel_order_from_orders_map(limit_orders_map_ptr, trading_pair_str, cancel_all=True)
            cancellation_results.extend(results)
        return cancellation_results

    cdef object c_cancel_order_from_orders_map(self,
                                               LimitOrders *orders_map,
                                               str trading_pair_str,
                                               bint cancel_all=False,
                                               str client_order_id=None):
        cdef:
            string cpp_trading_pair = trading_pair_str.encode("utf8")
            LimitOrdersIterator map_it = orders_map.find(cpp_trading_pair)
            SingleTradingPairLimitOrders *limit_orders_collection_ptr = NULL
            SingleTradingPairLimitOrdersIterator orders_it
            vector[SingleTradingPairLimitOrdersIterator] process_order_its
            const CPPLimitOrder *limit_order_ptr = NULL
            str limit_order_cid
            list cancellation_results = []
        try:
            if map_it == orders_map.end():
                return []

            limit_orders_collection_ptr = address(deref(map_it).second)
            orders_it = limit_orders_collection_ptr.begin()
            while orders_it != limit_orders_collection_ptr.end():
                limit_order_ptr = address(deref(orders_it))
                limit_order_cid = limit_order_ptr.getClientOrderID().decode("utf8")
                if (not cancel_all and limit_order_cid == client_order_id) or cancel_all:
                    process_order_its.push_back(orders_it)
                inc(orders_it)

            for orders_it in process_order_its:
                limit_order_ptr = address(deref(orders_it))
                limit_order_cid = limit_order_ptr.getClientOrderID().decode("utf8")
                delete_success = self.c_delete_limit_order(orders_map, address(map_it), orders_it)
                cancellation_results.append(CancellationResult(limit_order_cid,
                                                               delete_success))
                self.c_trigger_event(self.MARKET_ORDER_CANCELED_EVENT_TAG,
                                     OrderCancelledEvent(self._current_timestamp,
                                                         limit_order_cid)
                                     )
            return cancellation_results
        except Exception as err:
            self.logger().error(f"Error canceling order.", exc_info=True)

    cdef c_cancel(self, str trading_pair_str, str client_order_id):
        cdef:
            string cpp_trading_pair = trading_pair_str.encode("utf8")
            string cpp_client_order_id = client_order_id.encode("utf8")
            str trade_type = client_order_id.split("://")[0]
            bint is_maker_buy = trade_type.upper() == "BUY"
            LimitOrders *limit_orders_map_ptr = (address(self._bid_limit_orders)
                                                 if is_maker_buy
                                                 else address(self._ask_limit_orders))
        self.c_cancel_order_from_orders_map(limit_orders_map_ptr, trading_pair_str, False, client_order_id)

    cdef object c_get_fee(self,
                          str base_asset,
                          str quote_asset,
                          object order_type,
                          object order_side,
                          object amount,
                          object price,
                          object is_maker = None):
        return build_trade_fee(
            self.name,
            is_maker=is_maker if is_maker is not None else order_type in [OrderType.LIMIT, OrderType.LIMIT_MAKER],
            base_currency=base_asset,
            quote_currency=quote_asset,
            order_type=order_type,
            order_side=order_side,
            amount=amount,
            price=price,
        )

    cdef OrderBook c_get_order_book(self, str trading_pair):
        if trading_pair not in self._trading_pairs:
            raise ValueError(f"No order book exists for '{trading_pair}'.")
        trading_pair = self._target_market.convert_to_exchange_trading_pair(trading_pair)
        return self._order_book_tracker.order_books[trading_pair]

    cdef object c_get_order_price_quantum(self, str trading_pair, object price):
        cdef:
            QuantizationParams q_params
        if trading_pair in self._quantization_params:
            q_params = self._quantization_params[trading_pair]
            decimals_quantum = Decimal(f"1e-{q_params.price_decimals}")
            if price.is_finite() and price > s_decimal_0:
                precision_quantum = Decimal(f"1e{math.ceil(math.log10(price)) - q_params.price_precision}")
            else:
                precision_quantum = Decimal(0)
            return max(precision_quantum, decimals_quantum)
        else:
            return Decimal(f"1e-10")

    def get_order_price_quantum(self, trading_pair: str, price: Decimal) -> Decimal:
        return self.c_get_order_price_quantum(trading_pair, price)

    cdef object c_get_order_size_quantum(self,
                                         str trading_pair,
                                         object order_size):
        cdef:
            QuantizationParams q_params
        if trading_pair in self._quantization_params:
            q_params = self._quantization_params[trading_pair]
            decimals_quantum = Decimal(f"1e-{q_params.order_size_decimals}")
            if order_size.is_finite() and order_size > s_decimal_0:
                precision_quantum = Decimal(f"1e{math.ceil(math.log10(order_size)) - q_params.order_size_precision}")
            else:
                precision_quantum = Decimal(0)
            return max(precision_quantum, decimals_quantum)
        else:
            return Decimal(f"1e-7")

    cdef object c_quantize_order_price(self,
                                       str trading_pair,
                                       object price):
        price = Decimal('%.7g' % price)  # hard code to round to 8 significant digits
        price_quantum = self.c_get_order_price_quantum(trading_pair, price)
        return (price // price_quantum) * price_quantum

    cdef object c_quantize_order_amount(self,
                                        str trading_pair,
                                        object amount,
                                        object price=s_decimal_0):
        amount = amount.quantize(Decimal('1e-7'), rounding=ROUND_DOWN)
        if amount <= 1e-7:
            amount = Decimal("0")
        order_size_quantum = self.c_get_order_size_quantum(trading_pair, amount)
        return (amount // order_size_quantum) * order_size_quantum

    def get_available_balance(self, currency: str) -> Decimal:
        return self.c_get_available_balance(currency)

    def get_all_balances(self) -> Dict[str, Decimal]:
        return self._account_balances.copy()

    # <editor-fold desc="Python wrapper for cdef functions">
    def match_trade_to_limit_orders(self, event_object: OrderBookTradeEvent):
        self.c_match_trade_to_limit_orders(event_object)

    def set_balance(self, currency: str, balance: Decimal):
        self.c_set_balance(currency, balance)
    # </editor-fold>

    def get_price(self, trading_pair: str, is_buy: bool) -> Decimal:
        return self.c_get_price(trading_pair, is_buy)

    def buy(self, trading_pair: str, amount: Decimal, order_type=OrderType.MARKET,
            price: Decimal = s_decimal_0, **kwargs) -> str:
        return self.c_buy(trading_pair, amount, order_type, price, kwargs)

    def sell(self, trading_pair: str, amount: Decimal, order_type=OrderType.MARKET,
             price: Decimal = s_decimal_0, **kwargs) -> str:
        return self.c_sell(trading_pair, amount, order_type, price, kwargs)

    def cancel(self, trading_pair: str, client_order_id: str):
        return self.c_cancel(trading_pair, client_order_id)

    def get_fee(self,
                base_currency: str,
                quote_currency: str,
                order_type: OrderType,
                order_side: TradeType,
                amount: Decimal,
                price: Decimal = s_decimal_0,
                is_maker: Optional[bool] = None):
        return self.c_get_fee(base_currency, quote_currency, order_type, order_side, amount, price, is_maker)

    def get_order_book(self, trading_pair: str) -> OrderBook:
        return self.c_get_order_book(trading_pair)

    def get_maker_order_type(self):
        return OrderType.LIMIT

    def get_taker_order_type(self):
        return OrderType.LIMIT

    async def trigger_event_async(self,
                                  event_tag,
                                  event):
        await asyncio.sleep(0.01)
        self.c_trigger_event(event_tag, event)



from decimal import Decimal
from typing import TYPE_CHECKING, Dict, Optional

from hummingbot.connector.utils import split_hb_trading_pair
from hummingbot.core.rate_oracle.sources.rate_source_base import RateSourceBase
from hummingbot.core.utils import async_ttl_cache

if TYPE_CHECKING:
    from hummingbot.connector.derivative.aevo_perpetual.aevo_perpetual_derivative import AevoPerpetualDerivative


class AevoRateSource(RateSourceBase):
    def __init__(self):
        super().__init__()
        self._exchange: Optional[AevoPerpetualDerivative] = None

    @property
    def name(self) -> str:
        return "aevo_perpetual"

    @async_ttl_cache(ttl=30, maxsize=1)
    async def get_prices(self, quote_token: Optional[str] = None) -> Dict[str, Decimal]:
        self._ensure_exchange()
        results = {}

        try:
            pairs_prices = await self._exchange.get_all_pairs_prices()

            for pair_price in pairs_prices:
                pair = pair_price["symbol"]

                try:
                    trading_pair = await self._exchange.trading_pair_associated_to_exchange_symbol(symbol=pair)
                except KeyError:
                    continue

                if quote_token is not None:
                    _, quote = split_hb_trading_pair(trading_pair=trading_pair)

                    if quote != quote_token:
                        continue

                price = pair_price["price"]

                if price is not None:
                    results[trading_pair] = Decimal(price)
        except Exception:
            self.logger().exception(
                msg="Unexpected error while retrieving rates from Aevo. Check the log file for more info.",
            )
        return results

    def _ensure_exchange(self):
        if self._exchange is None:
            self._exchange = self._build_aevo_connector_without_private_keys()

    @staticmethod
    def _build_aevo_connector_without_private_keys() -> 'AevoPerpetualDerivative':
        from hummingbot.connector.derivative.aevo_perpetual.aevo_perpetual_derivative import AevoPerpetualDerivative

        return AevoPerpetualDerivative(
            aevo_perpetual_api_key="",
            aevo_perpetual_api_secret="",
            aevo_perpetual_signing_key="",
            aevo_perpetual_account_address="",
            trading_pairs=[],
            trading_required=False,
        )



# distutils: language=c++

from .LimitOrder cimport LimitOrder as CPPLimitOrder


cdef class LimitOrder:
    cdef:
        CPPLimitOrder _cpp_limit_order
    cdef long long c_age(self)
    cdef long long c_age_til(self, long long start_timestamp)


cdef LimitOrder c_create_limit_order_from_cpp_limit_order(const CPPLimitOrder cpp_limit_order)



import asyncio
import logging
import time
from decimal import Decimal
from typing import Dict, List, Optional, Tuple

import pandas as pd

from hummingbot.client.config.config_helpers import (
    ClientConfigAdapter,
    api_keys_from_connector_config_map,
    get_connector_class,
)
from hummingbot.client.settings import AllConnectorSettings
from hummingbot.connector.connector_base import ConnectorBase
from hummingbot.core.data_type.common import GroupedSetDict, LazyDict, PriceType, TradeType
from hummingbot.core.data_type.order_book_query_result import OrderBookQueryResult
from hummingbot.core.gateway.gateway_http_client import GatewayHttpClient
from hummingbot.core.rate_oracle.rate_oracle import RateOracle
from hummingbot.core.utils.async_utils import safe_ensure_future
from hummingbot.data_feed.candles_feed.candles_factory import CandlesFactory
from hummingbot.data_feed.candles_feed.data_types import CandlesConfig
from hummingbot.logger import HummingbotLogger
from hummingbot.strategy_v2.executors.data_types import ConnectorPair


class MarketDataProvider:
    _logger: Optional[HummingbotLogger] = None
    gateway_price_provider_by_chain: Dict = {
        "ethereum": "uniswap/router",
        "solana": "jupiter/router",
    }

    @classmethod
    def logger(cls) -> HummingbotLogger:
        if cls._logger is None:
            cls._logger = logging.getLogger(__name__)
        return cls._logger

    def __init__(self,
                 connectors: Dict[str, ConnectorBase],
                 rates_update_interval: int = 60):
        self.candles_feeds = {}  # Stores instances of candle feeds
        self.connectors = connectors  # Stores instances of connectors
        self._rates_update_task = None
        self._rates_update_interval = rates_update_interval
        self._rates = {}
        self._non_trading_connectors = LazyDict[str, ConnectorBase](self._create_non_trading_connector)
        self._non_trading_connectors_started: Dict[str, bool] = {}  # Track which connectors have been started
        self._rates_required = GroupedSetDict[str, ConnectorPair]()
        self.conn_settings = AllConnectorSettings.get_connector_settings()

    def stop(self):
        for candle_feed in self.candles_feeds.values():
            candle_feed.stop()
        if self._rates_update_task:
            self._rates_update_task.cancel()
            self._rates_update_task = None
        self.candles_feeds.clear()
        self._rates_required.clear()
        # Stop non-trading connectors that were started for public data access
        for connector in self._non_trading_connectors.values():
            safe_ensure_future(connector.stop_network())
        self._non_trading_connectors.clear()
        self._non_trading_connectors_started.clear()

    @property
    def ready(self) -> bool:
        all_connectors_running = all(connector.ready for connector in self.connectors.values())
        all_candles_feeds_running = all(feed.ready for feed in self.candles_feeds.values())
        return all_connectors_running and all_candles_feeds_running

    def time(self):
        return time.time()

    def initialize_rate_sources(self, connector_pairs: List[ConnectorPair]):
        """
        Initializes a rate source based on the given connector pair.
        :param connector_pairs: List[ConnectorPair]
        """
        for connector_pair in connector_pairs:
            self._rates_required.add_or_update(connector_pair.connector_name, connector_pair)
        if not self._rates_update_task:
            self._rates_update_task = safe_ensure_future(self.update_rates_task())

    def remove_rate_sources(self, connector_pairs: List[ConnectorPair]):
        """
        Removes rate sources for the given connector pairs.
        :param connector_pairs: List[ConnectorPair]
        """
        for connector_pair in connector_pairs:
            self._rates_required.remove(connector_pair.connector_name, connector_pair)

        # Stop the rates update task if no more rates are required
        if len(self._rates_required) == 0 and self._rates_update_task:
            self._rates_update_task.cancel()
            self._rates_update_task = None

    async def update_rates_task(self):
        """
        Updates the rates for all rate sources.
        """
        try:
            while True:
                # Exit if no more rates to update
                if len(self._rates_required) == 0:
                    break

                rate_oracle = RateOracle.get_instance()

                # Separate gateway and non-gateway connectors
                gateway_tasks = []
                gateway_task_metadata = []  # Store (connector_pair, trading_pair) for each task
                non_gateway_connectors = {}

                for connector, connector_pairs in self._rates_required.items():
                    # Detect gateway connectors: either new format (contains "/") or old format (contains "gateway")
                    is_gateway = "/" in connector or "gateway" in connector

                    if is_gateway:
                        gateway_client = GatewayHttpClient.get_instance()

                        for connector_pair in connector_pairs:
                            try:
                                base, quote = connector_pair.trading_pair.split("-")

                                # Parse connector format to extract chain/network
                                if "/" in connector:
                                    # New format: "jupiter/router" or "uniswap/amm"
                                    # Need to get chain/network from the connector instance or Gateway
                                    chain, network, error = await gateway_client.get_connector_chain_network(connector)
                                    if error:
                                        self.logger().warning(f"Failed to get chain/network for {connector}: {error}")
                                        continue
                                    connector_name = connector  # Use the connector as-is for new format
                                else:
                                    # Old format: "gateway_chain-network"
                                    gateway, chain_network = connector.split("_", 1)
                                    chain, network = chain_network.split("-", 1)
                                    connector_name = self.gateway_price_provider_by_chain.get(chain)
                                    if not connector_name:
                                        self.logger().warning(f"No gateway price provider found for chain {chain}")
                                        continue

                                task = gateway_client.get_price(
                                    chain=chain,
                                    network=network,
                                    connector=connector_name,
                                    base_asset=base,
                                    quote_asset=quote,
                                    amount=Decimal("1"),
                                    side=TradeType.SELL
                                )
                                gateway_tasks.append(task)
                                gateway_task_metadata.append((connector_pair, connector_pair.trading_pair))

                            except Exception as e:
                                self.logger().warning(f"Error preparing price request for {connector_pair.trading_pair}: {e}")
                                continue
                    else:
                        # Non-gateway connector
                        non_gateway_connectors[connector] = connector_pairs

                # Gather ALL gateway tasks at once for maximum parallelization
                if gateway_tasks:
                    try:
                        results = await asyncio.gather(*gateway_tasks, return_exceptions=True)
                        for (connector_pair, trading_pair), rate in zip(gateway_task_metadata, results):
                            if isinstance(rate, Exception):
                                self.logger().error(f"Error fetching price for {trading_pair}: {rate}")
                            elif rate and "price" in rate:
                                rate_oracle.set_price(trading_pair, Decimal(rate["price"]))
                    except Exception as e:
                        self.logger().error(f"Error fetching gateway prices: {e}", exc_info=True)

                # Process non-gateway connectors
                for connector, connector_pairs in non_gateway_connectors.items():
                    try:
                        connector_instance = self._non_trading_connectors[connector]
                        prices = await self._safe_get_last_traded_prices(
                            connector=connector_instance,
                            trading_pairs=[pair.trading_pair for pair in connector_pairs])
                        for pair, rate in prices.items():
                            rate_oracle.set_price(pair, rate)
                    except Exception as e:
                        self.logger().error(f"Error fetching prices from {connector}: {e}", exc_info=True)

                await asyncio.sleep(self._rates_update_interval)
        except asyncio.CancelledError:
            raise
        finally:
            self._rates_update_task = None

    def initialize_candles_feed(self, config: CandlesConfig):
        """
        Initializes a candle feed based on the given configuration.
        :param config: CandlesConfig
        """
        self.get_candles_feed(config)

    def initialize_candles_feed_list(self, config_list: List[CandlesConfig]):
        """
        Initializes a list of candle feeds based on the given configurations.
        :param config_list: List[CandlesConfig]
        """
        for config in config_list:
            self.get_candles_feed(config)

    def get_candles_feed(self, config: CandlesConfig):
        """
        Retrieves or creates and starts a candle feed based on the given configuration.
        If an existing feed has a higher or equal max_records, it is reused.
        :param config: CandlesConfig
        :return: Candle feed instance.
        """
        key = self._generate_candle_feed_key(config)
        existing_feed = self.candles_feeds.get(key)

        if existing_feed and existing_feed.max_records >= config.max_records:
            # Existing feed is sufficient, return it
            return existing_feed
        else:
            # Stop the existing feed if it exists before creating a new one
            if existing_feed and hasattr(existing_feed, 'stop'):
                existing_feed.stop()

            # Create a new feed with updated max_records
            candle_feed = CandlesFactory.get_candle(config)
            self.candles_feeds[key] = candle_feed
            if hasattr(candle_feed, 'start'):
                candle_feed.start()
            return candle_feed

    @staticmethod
    def _generate_candle_feed_key(config: CandlesConfig) -> str:
        """
        Generates a unique key for a candle feed based on its configuration.
        :param config: CandlesConfig
        :return: Unique key as a string.
        """
        return f"{config.connector}_{config.trading_pair}_{config.interval}"

    def stop_candle_feed(self, config: CandlesConfig):
        """
        Stops a candle feed based on the given configuration.
        :param config: CandlesConfig
        """
        key = self._generate_candle_feed_key(config)
        candle_feed = self.candles_feeds.get(key)
        if candle_feed and hasattr(candle_feed, 'stop'):
            candle_feed.stop()
            del self.candles_feeds[key]

    def get_connector(self, connector_name: str) -> ConnectorBase:
        """
        Retrieves a connector instance based on the given name.
        :param connector_name: str
        :return: ConnectorBase
        """
        connector = self.connectors.get(connector_name)
        if not connector:
            raise ValueError(f"Connector {connector_name} not found.")
        return connector

    def get_connector_with_fallback(self, connector_name: str) -> ConnectorBase:
        """
        Retrieves a connector instance with fallback to non-trading connector.
        Prefers existing connected connector with API keys if available,
        otherwise creates a non-trading connector for public data access.
        :param connector_name: str
        :return: ConnectorBase
        """
        # Try to get existing connector first (has API keys)
        connector = self.connectors.get(connector_name)
        if connector:
            return connector

        # Fallback to non-trading connector for public data
        return self.get_non_trading_connector(connector_name)

    def get_non_trading_connector(self, connector_name: str):
        """
        Retrieves a non-trading connector from cache or creates one if not exists.
        Uses the _non_trading_connectors cache to avoid creating multiple instances.
        :param connector_name: str
        :return: ConnectorBase
        """
        return self._non_trading_connectors[connector_name]

    def _create_non_trading_connector(self, connector_name: str):
        """
        Creates a new non-trading connector instance.
        This is the factory method used by the LazyDict cache.
        Note: The connector is NOT started automatically. Call _ensure_non_trading_connector_started()
        to start it with at least one trading pair.
        :param connector_name: str
        :return: ConnectorBase
        """
        conn_setting = self.conn_settings.get(connector_name)
        if conn_setting is None:
            self.logger().error(f"Connector {connector_name} not found")
            raise ValueError(f"Connector {connector_name} not found")

        init_params = conn_setting.conn_init_parameters(
            trading_pairs=[],
            trading_required=False,
            api_keys=self.get_connector_config_map(connector_name),
        )
        connector_class = get_connector_class(connector_name)
        connector = connector_class(**init_params)
        return connector

    async def _ensure_non_trading_connector_started(
        self, connector: ConnectorBase, connector_name: str, trading_pair: str
    ) -> bool:
        """
        Ensures a non-trading connector is started with at least one trading pair.
        This is needed because exchanges like Binance close WebSocket connections
        that have no subscriptions.

        :param connector: ConnectorBase
        :param connector_name: str
        :param trading_pair: str - The first trading pair to subscribe to
        :return: True if connector was started or already running, False on error
        """
        if self._non_trading_connectors_started.get(connector_name, False):
            return True

        try:
            # Add the trading pair to the connector BEFORE starting the network
            # This ensures the WebSocket has something to subscribe to
            if trading_pair not in connector._trading_pairs:
                connector._trading_pairs.append(trading_pair)

            # Start the network - this will initialize order book tracker with the trading pair
            await connector.start_network()
            self._non_trading_connectors_started[connector_name] = True
            self.logger().info(f"Started non-trading connector: {connector_name} with initial pair {trading_pair}")

            # Wait for order book tracker to be ready
            max_wait = 30
            waited = 0
            tracker = connector.order_book_tracker
            while waited < max_wait:
                if tracker._order_book_stream_listener_task is not None:
                    # Give WebSocket time to establish connection
                    await asyncio.sleep(2.0)
                    break
                await asyncio.sleep(0.5)
                waited += 0.5

            return True
        except Exception as e:
            self.logger().error(f"Error starting non-trading connector {connector_name}: {e}")
            return False

    @staticmethod
    def get_connector_config_map(connector_name: str):
        connector_config = AllConnectorSettings.get_connector_config_keys(connector_name)
        if getattr(connector_config, "use_auth_for_public_endpoints", False):
            # Use real API keys for connectors that require auth for public endpoints
            api_keys = api_keys_from_connector_config_map(ClientConfigAdapter(connector_config))
        elif connector_config is not None:
            # Provide empty strings for all config keys (for public data access without auth)
            api_keys = {key: "" for key in connector_config.__class__.model_fields.keys() if key != "connector"}
        else:
            # No config found, return empty dict
            api_keys = {}
        return api_keys

    def get_balance(self, connector_name: str, asset: str):
        connector = self.get_connector(connector_name)
        return connector.get_balance(asset)

    def get_available_balance(self, connector_name: str, asset: str):
        connector = self.get_connector(connector_name)
        return connector.get_available_balance(asset)

    def get_order_book(self, connector_name: str, trading_pair: str):
        """
        Retrieves the order book for a trading pair from the specified connector.
        :param connector_name: str
        :param trading_pair: str
        :return: Order book instance.
        """
        connector = self.get_connector_with_fallback(connector_name)
        return connector.get_order_book(trading_pair)

    async def initialize_order_book(self, connector_name: str, trading_pair: str) -> bool:
        """
        Dynamically initializes order book for a trading pair on the specified connector.
        This subscribes to the order book WebSocket channel and starts tracking the pair.

        For perpetual connectors, this also initializes funding info and other perpetual-specific data.

        :param connector_name: str
        :param trading_pair: str
        :return: True if successful, False otherwise
        """
        connector = self.get_connector_with_fallback(connector_name)
        if not hasattr(connector, 'order_book_tracker'):
            self.logger().warning(f"Connector {connector_name} does not have order_book_tracker")
            return False

        # For non-trading connectors, ensure the network is started with this trading pair
        if connector_name not in self.connectors:
            if not self._non_trading_connectors_started.get(connector_name, False):
                # First time - start the connector with this trading pair as the initial subscription
                success = await self._ensure_non_trading_connector_started(
                    connector, connector_name, trading_pair
                )
                if not success:
                    return False
                # The trading pair was added during startup, so we're done
                # Wait for order book to be initialized
                await self._wait_for_order_book_initialized(connector, trading_pair)
                return True

        # Add trading pair dynamically via connector method
        return await connector.add_trading_pair(trading_pair)

    async def _wait_for_order_book_initialized(
        self, connector: ConnectorBase, trading_pair: str, timeout: float = 30.0
    ) -> bool:
        """
        Waits for an order book to be initialized for a trading pair.

        :param connector: ConnectorBase
        :param trading_pair: str
        :param timeout: Maximum time to wait in seconds
        :return: True if initialized, False if timeout
        """
        tracker = connector.order_book_tracker
        waited = 0
        interval = 0.5
        while waited < timeout:
            if trading_pair in tracker.order_books:
                ob = tracker.order_books[trading_pair]
                bids, asks = ob.snapshot
                if len(bids) > 0 and len(asks) > 0:
                    self.logger().info(f"Order book for {trading_pair} initialized successfully")
                    return True
            await asyncio.sleep(interval)
            waited += interval
        self.logger().warning(f"Timeout waiting for {trading_pair} order book to initialize")
        return False

    async def initialize_order_books(self, connector_name: str, trading_pairs: List[str]) -> Dict[str, bool]:
        """
        Dynamically initializes order books for multiple trading pairs in parallel.

        :param connector_name: str
        :param trading_pairs: List[str]
        :return: Dict mapping trading pair to success status
        """
        tasks = [self.initialize_order_book(connector_name, tp) for tp in trading_pairs]
        results = await asyncio.gather(*tasks, return_exceptions=True)
        return {
            tp: (result is True) if not isinstance(result, Exception) else False
            for tp, result in zip(trading_pairs, results)
        }

    async def remove_order_book(self, connector_name: str, trading_pair: str) -> bool:
        """
        Removes order book tracking for a trading pair from the specified connector.
        This unsubscribes from the WebSocket channel and stops tracking the pair.

        For perpetual connectors, this also cleans up funding info and other perpetual-specific data.

        :param connector_name: str
        :param trading_pair: str
        :return: True if successful, False otherwise
        """
        connector = self.get_connector_with_fallback(connector_name)
        if not hasattr(connector, 'order_book_tracker'):
            self.logger().warning(f"Connector {connector_name} does not have order_book_tracker")
            return False

        # Remove trading pair via connector method
        return await connector.remove_trading_pair(trading_pair)

    async def remove_order_books(self, connector_name: str, trading_pairs: List[str]) -> Dict[str, bool]:
        """
        Removes order book tracking for multiple trading pairs in parallel.

        :param connector_name: str
        :param trading_pairs: List[str]
        :return: Dict mapping trading pair to success status
        """
        tasks = [self.remove_order_book(connector_name, tp) for tp in trading_pairs]
        results = await asyncio.gather(*tasks, return_exceptions=True)
        return {
            tp: (result is True) if not isinstance(result, Exception) else False
            for tp, result in zip(trading_pairs, results)
        }

    def get_price_by_type(self, connector_name: str, trading_pair: str, price_type: PriceType = PriceType.MidPrice):
        """
        Retrieves the price for a trading pair from the specified connector.
        :param connector_name: str
        :param trading_pair: str
        :param price_type: str
        :return: Price instance.
        """
        connector = self.get_connector_with_fallback(connector_name)
        return connector.get_price_by_type(trading_pair, price_type)

    def get_funding_info(self, connector_name: str, trading_pair: str):
        """
        Retrieves the funding rate for a trading pair from the specified connector.
        :param connector_name: str
        :param trading_pair: str
        :return: Funding rate.
        """
        connector = self.get_connector_with_fallback(connector_name)
        return connector.get_funding_info(trading_pair)

    def get_candles_df(self, connector_name: str, trading_pair: str, interval: str, max_records: int = 500):
        """
        Retrieves the candles for a trading pair from the specified connector.
        :param connector_name: str
        :param trading_pair: str
        :param interval: str
        :param max_records: int
        :return: Candles dataframe.
        """
        candles = self.get_candles_feed(CandlesConfig(
            connector=connector_name,
            trading_pair=trading_pair,
            interval=interval,
            max_records=max_records,
        ))
        return candles.candles_df.iloc[-max_records:]

    async def get_historical_candles_df(self, connector_name: str, trading_pair: str, interval: str,
                                        start_time: Optional[int] = None, end_time: Optional[int] = None,
                                        max_records: Optional[int] = None, max_cache_records: int = 10000):
        """
        Retrieves historical candles with intelligent caching and partial fetch optimization.

        :param connector_name: str
        :param trading_pair: str
        :param interval: str
        :param start_time: Start timestamp in seconds (optional)
        :param end_time: End timestamp in seconds (optional)
        :param max_records: Maximum number of records to return (optional)
        :param max_cache_records: Maximum records to keep in cache for efficiency
        :return: Candles dataframe for the requested range
        """
        import time

        from hummingbot.data_feed.candles_feed.data_types import HistoricalCandlesConfig

        # Set default end_time to current time if not provided
        if end_time is None:
            end_time = int(time.time())

        # Calculate start_time based on max_records if not provided
        if start_time is None and max_records is not None:
            # Get interval in seconds to calculate approximate start time
            candles_feed = self.get_candles_feed(CandlesConfig(
                connector=connector_name,
                trading_pair=trading_pair,
                interval=interval,
                max_records=min(100, max_records)  # Small initial fetch to get interval info
            ))
            interval_seconds = candles_feed.interval_in_seconds
            start_time = end_time - (max_records * interval_seconds)

        if start_time is None:
            # Fallback to regular method if no time range specified
            return self.get_candles_df(connector_name, trading_pair, interval, max_records or 500)

        # Get or create candles feed with extended cache
        candles_feed = self.get_candles_feed(CandlesConfig(
            connector=connector_name,
            trading_pair=trading_pair,
            interval=interval,
            max_records=max_cache_records
        ))

        # Check if we have cached data and what range it covers
        current_df = candles_feed.candles_df

        if len(current_df) > 0:
            cached_start = int(current_df['timestamp'].iloc[0])
            cached_end = int(current_df['timestamp'].iloc[-1])

            # Check if requested range is completely covered by cache
            if start_time >= cached_start and end_time <= cached_end:
                # Filter existing data for requested range
                filtered_df = current_df[
                    (current_df['timestamp'] >= start_time) &
                    (current_df['timestamp'] <= end_time)
                ]
                return filtered_df.iloc[-max_records:] if max_records else filtered_df

            # Partial cache hit - determine what additional data we need
            fetch_start = min(start_time, cached_start)
            fetch_end = max(end_time, cached_end)

            # If the extended range is too large, limit it
            max_fetch_range = max_cache_records * candles_feed.interval_in_seconds
            if (fetch_end - fetch_start) > max_fetch_range:
                # Prioritize the requested range
                if start_time < cached_start:
                    fetch_start = max(start_time, fetch_end - max_fetch_range)
                else:
                    fetch_end = min(end_time, fetch_start + max_fetch_range)
        else:
            # No cached data - fetch requested range with some buffer
            buffer_records = min(max_cache_records // 4, 1000)  # 25% buffer or 1000 records max
            interval_seconds = candles_feed.interval_in_seconds
            buffer_time = buffer_records * interval_seconds

            fetch_start = start_time - buffer_time
            fetch_end = end_time

        # Fetch historical data
        try:
            historical_config = HistoricalCandlesConfig(
                connector_name=connector_name,
                trading_pair=trading_pair,
                interval=interval,
                start_time=fetch_start,
                end_time=fetch_end
            )

            new_df = await candles_feed.get_historical_candles(historical_config)

            if len(new_df) > 0:
                # Merge with existing data if any
                if len(current_df) > 0:
                    combined_df = pd.concat([current_df, new_df], ignore_index=True)
                    # Remove duplicates and sort
                    combined_df = combined_df.drop_duplicates(subset=['timestamp'])
                    combined_df = combined_df.sort_values('timestamp')

                    # Limit cache size
                    if len(combined_df) > max_cache_records:
                        # Keep most recent records
                        combined_df = combined_df.iloc[-max_cache_records:]

                    # Update the candles feed cache
                    candles_feed._candles.clear()
                    for _, row in combined_df.iterrows():
                        candles_feed._candles.append(row.values)
                else:
                    # Update the candles feed cache with new data
                    candles_feed._candles.clear()
                    for _, row in new_df.iloc[-max_cache_records:].iterrows():
                        candles_feed._candles.append(row.values)

                # Return filtered data for requested range
                final_df = candles_feed.candles_df
                filtered_df = final_df[
                    (final_df['timestamp'] >= start_time) &
                    (final_df['timestamp'] <= end_time)
                ]
                return filtered_df.iloc[-max_records:] if max_records else filtered_df

        except Exception as e:
            self.logger().warning(f"Error fetching historical candles: {e}. Falling back to regular method.")

        # Fallback to existing method if historical fetch fails
        return self.get_candles_df(connector_name, trading_pair, interval, max_records or 500)

    def get_trading_pairs(self, connector_name: str):
        """
        Retrieves the trading pairs from the specified connector.
        :param connector_name: str
        :return: List of trading pairs.
        """
        connector = self.get_connector_with_fallback(connector_name)
        return connector.trading_pairs

    def get_trading_rules(self, connector_name: str, trading_pair: str):
        """
        Retrieves the trading rules from the specified connector.
        :param connector_name: str
        :return: Trading rules.
        """
        connector = self.get_connector_with_fallback(connector_name)
        return connector.trading_rules[trading_pair]

    def quantize_order_price(self, connector_name: str, trading_pair: str, price: Decimal):
        connector = self.get_connector_with_fallback(connector_name)
        return connector.quantize_order_price(trading_pair, price)

    def quantize_order_amount(self, connector_name: str, trading_pair: str, amount: Decimal):
        connector = self.get_connector_with_fallback(connector_name)
        return connector.quantize_order_amount(trading_pair, amount)

    def get_price_for_volume(self, connector_name: str, trading_pair: str, volume: float,
                             is_buy: bool) -> OrderBookQueryResult:
        """
        Gets the price for a specified volume on the order book.

        :param connector_name: The name of the connector.
        :param trading_pair: The trading pair for which to retrieve the data.
        :param volume: The volume for which to find the price.
        :param is_buy: True if buying, False if selling.
        :return: OrderBookQueryResult containing the result of the query.
        """
        connector = self.get_connector_with_fallback(connector_name)
        order_book = connector.get_order_book(trading_pair)
        return order_book.get_price_for_volume(is_buy, volume)

    def get_order_book_snapshot(self, connector_name, trading_pair) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """
        Retrieves the order book snapshot for a trading pair from the specified connector, as a tuple of bid and ask in
        DataFrame format.
        :param connector_name: str
        :param trading_pair: str
        :return: Tuple of bid and ask in DataFrame format.
        """
        connector = self.get_connector_with_fallback(connector_name)
        order_book = connector.get_order_book(trading_pair)
        return order_book.snapshot

    def get_price_for_quote_volume(self, connector_name: str, trading_pair: str, quote_volume: float,
                                   is_buy: bool) -> OrderBookQueryResult:
        """
        Gets the price for a specified quote volume on the order book.

        :param connector_name: The name of the connector.
        :param trading_pair: The trading pair for which to retrieve the data.
        :param quote_volume: The quote volume for which to find the price.
        :param is_buy: True if buying, False if selling.
        :return: OrderBookQueryResult containing the result of the query.
        """
        connector = self.get_connector_with_fallback(connector_name)
        order_book = connector.get_order_book(trading_pair)
        return order_book.get_price_for_quote_volume(is_buy, quote_volume)

    def get_volume_for_price(self, connector_name: str, trading_pair: str, price: float,
                             is_buy: bool) -> OrderBookQueryResult:
        """
        Gets the volume for a specified price on the order book.

        :param connector_name: The name of the connector.
        :param trading_pair: The trading pair for which to retrieve the data.
        :param price: The price for which to find the volume.
        :param is_buy: True if buying, False if selling.
        :return: OrderBookQueryResult containing the result of the query.
        """
        connector = self.get_connector_with_fallback(connector_name)
        order_book = connector.get_order_book(trading_pair)
        return order_book.get_volume_for_price(is_buy, price)

    def get_quote_volume_for_price(self, connector_name: str, trading_pair: str, price: float,
                                   is_buy: bool) -> OrderBookQueryResult:
        """
        Gets the quote volume for a specified price on the order book.

        :param connector_name: The name of the connector.
        :param trading_pair: The trading pair for which to retrieve the data.
        :param price: The price for which to find the quote volume.
        :param is_buy: True if buying, False if selling.
        :return: OrderBookQueryResult containing the result of the query.
        """
        connector = self.get_connector_with_fallback(connector_name)
        order_book = connector.get_order_book(trading_pair)
        return order_book.get_quote_volume_for_price(is_buy, price)

    def get_vwap_for_volume(self, connector_name: str, trading_pair: str, volume: float,
                            is_buy: bool) -> OrderBookQueryResult:
        """
        Gets the VWAP (Volume Weighted Average Price) for a specified volume on the order book.

        :param connector_name: The name of the connector.
        :param trading_pair: The trading pair for which to retrieve the data.
        :param volume: The volume for which to calculate the VWAP.
        :param is_buy: True if buying, False if selling.
        :return: OrderBookQueryResult containing the result of the query.
        """
        connector = self.get_connector_with_fallback(connector_name)
        order_book = connector.get_order_book(trading_pair)
        return order_book.get_vwap_for_volume(is_buy, volume)

    def get_rate(self, pair: str) -> Decimal:
        """
        Finds a conversion rate for a given trading pair, this can be direct or indirect prices as
        long as it can find a route to achieve this.

        :param pair: A trading pair, e.g. BTC-USDT
        :return A conversion rate
        """
        return RateOracle.get_instance().get_pair_rate(pair)

    async def _safe_get_last_traded_prices(self, connector, trading_pairs, timeout=5):
        try:
            tasks = [self._safe_get_last_traded_price(connector, trading_pair) for trading_pair in trading_pairs]
            prices = await asyncio.wait_for(asyncio.gather(*tasks), timeout=timeout)
            return {pair: Decimal(rate) for pair, rate in zip(trading_pairs, prices)}
        except Exception as e:
            logging.error(f"Error getting last traded prices in connector {connector} for trading pairs {trading_pairs}: {e}")
            return {}

    async def _safe_get_last_traded_price(self, connector, trading_pair):
        try:
            last_traded = await connector._get_last_traded_price(trading_pair=trading_pair)
            return Decimal(last_traded)
        except Exception as e:
            logging.error(f"Error getting last traded price in connector {connector} for trading pair {trading_pair}: {e}")
            return Decimal(0)



# distutils: language=c++

from libc.stdint cimport int64_t
from libcpp.unordered_map cimport unordered_map
from libcpp.unordered_set cimport unordered_set
from libcpp.utility cimport pair
from hummingbot.core.PyRef cimport PyRef
from hummingbot.core.event.event_listener cimport EventListener

ctypedef unordered_set[PyRef] EventListenersCollection
ctypedef unordered_set[PyRef].iterator EventListenersIterator
ctypedef unordered_map[int64_t, EventListenersCollection] Events
ctypedef unordered_map[int64_t, EventListenersCollection].iterator EventsIterator
ctypedef pair[int64_t, EventListenersCollection] EventsPair


cdef class PubSub:
    cdef:
        Events _events
        object __weakref__

    cdef c_log_exception(self, int64_t event_tag, object arg)
    cdef c_add_listener(self, int64_t event_tag, EventListener listener)
    cdef c_remove_listener(self, int64_t event_tag, EventListener listener)
    cdef c_remove_dead_listeners(self, int64_t event_tag)
    cdef c_get_listeners(self, int64_t event_tag)
    cdef c_trigger_event(self, int64_t event_tag, object arg)
